from subprocess import Popen
import os
from time import sleep

import paramiko

from brome.core.utils import (
    string_to_filename,
    devnull,
    create_dir_if_doesnt_exist
)
from brome.runner.base_instance import BaseInstance
from brome.core.settings import BROME_CONFIG


class EC2Instance(BaseInstance):
    """EC2 instance

    Attributes:
        runner (object)
        browser_config (object)
        index (int)
    """

    def __init__(self, runner, browser_config, index):
        self.runner = runner
        self.browser_config = browser_config
        self.index = index

    def get_ip(self):
        """Return the ip address of the node
        """

        return self.private_ip

    def scp_file_remote_to_local(self, remote_path, local_path):
        """Scp a remote file to local

        Args:
            remote_path (str)
            local_path (str)
        """

        scp_command = [
            'scp',
            '-o',
            'StrictHostKeyChecking=no',
            '-i',
            self.browser_config.get('ssh_key_path'),
            '%s@%s:"%s"' %
            (
                self.browser_config.get('username'),
                self.get_ip(),
                remote_path
            ),
            local_path
        ]
        self.info_log(
            "executing command: %s" %
            ' '.join(scp_command)
        )
        p = Popen(scp_command)
        p.wait()

    def execute_command(self, command):
        """Execute a command on the node

        Args:
            command (str)
        """

        self.info_log("executing command: %s" % command)

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            k = paramiko.RSAKey.from_private_key_file(
                self.browser_config.get('ssh_key_path')
            )
            ssh.connect(
                self.private_ip,
                username=self.browser_config.get('username'),
                pkey=k
            )

            sleep_time = 0.1
            stdout = []
            stderr = []

            ssh_transport = ssh.get_transport()
            channel = ssh_transport.open_session()
            channel.setblocking(0)
            channel.exec_command(command)

            while True:

                while channel.recv_ready():
                    stdout.append(channel.recv(1000))

                while channel.recv_stderr_ready():
                    stderr.append(channel.recv_stderr(1000))

                if channel.exit_status_ready():
                    break

                sleep(sleep_time)

            # ret = channel.recv_exit_status()
            ssh_transport.close()

            ssh.close()

            return b''.join(stdout), b''.join(stderr)

        except Exception as e:
            msg = "Execute_command exception: %s" % str(e)
            self.error_log(msg)
            raise Exception(msg)

    def startup(self):
        """Startup the ec2 instance
        """
        import boto.ec2

        if not self.browser_config.get('launch'):
            self.warning_log("Skipping launch")
            return True

        self.info_log("Starting up")

        instance = None
        try:

            # KEY NAME
            key_name = self.browser_config.get(
                "ssh_key_path"
            ).split(os.sep)[-1][:-4]
            # SECURITY GROUP

            if type(self.browser_config.get("security_group_ids")) == str:
                security_group_ids = [
                    self.browser_config.get("security_group_ids")
                ]

            elif type(self.browser_config.get("security_group_ids")) == list:
                security_group_ids = self.browser_config.get(
                    "security_group_ids"
                )

            else:
                msg = "The config security_group_ids must be a string or a list of string"  # noqa
                self.critial_log(msg)
                raise Exception(msg)

            # LAUNCH INSTANCE
            ec2 = boto.ec2.connect_to_region(self.browser_config.get("region"))
            reservation = ec2.run_instances(
                    self.browser_config.get('amiid'),
                    key_name=key_name,
                    instance_type=self.browser_config.get("instance_type"),
                    security_group_ids=security_group_ids
            )

            wait_after_instance_launched = BROME_CONFIG['ec2']['wait_after_instance_launched']  # noqa
            if wait_after_instance_launched:
                self.info_log(
                    "Waiting after instance launched: %s seconds..." %
                    wait_after_instance_launched
                )
                sleep(wait_after_instance_launched)

            else:
                self.warning_log("Skipping waiting after instance launched")

            try:
                instance = reservation.instances[0]

            except Exception as e:
                self.critical_log(
                    'Instance reservation exception: %s' % str(e)
                )
                raise

            self.instance_id = instance.id

            self.info_log('Waiting for the instance to start...')

            for i in range(60*5):
                try:
                    status = instance.update()
                    if status == 'running':
                        break

                    else:
                        sleep(1)
                except Exception as e:
                    self.error_log(
                        'Exception while wait pending: %s' % str(e)
                    )
                    sleep(1)

            # Wait until instance is running
            status = instance.update()
            if status == 'running':
                instance.add_tag(
                    "Name",
                    "%s-selenium-node-%s-%s" %
                    (
                        self.browser_config.get('platform'),
                        self.browser_config.get('browserName'),
                        self.index
                    )
                )

                self.info_log(
                    "New instance (%s) public ip (%s) private ip (%s)" % (
                        instance.id,
                        instance.ip_address,
                        instance.private_ip_address
                    )
                )
            else:
                msg = "Instance status is %s and should be (running)" % status
                self.error_log(msg)
                raise Exception(msg)

            if BROME_CONFIG['ec2']['wait_until_system_and_instance_check_performed']:  # noqa
                check_successful = False

                for i in range(5*60):

                    try:

                        if not i % 60:
                            if not type(status) == str:
                                self.info_log(
                                    'System_status: %s, instance_status: %s' %
                                    (
                                        status.system_status,
                                        status.instance_status
                                    )
                                )

                        status = ec2.get_all_instance_status(
                            instance_ids=[instance.id]
                        )[0]
                        if status.system_status.status == u'ok' and status.instance_status.status == u'ok':  # noqa

                            self.info_log('system_status: %s, instance_status: %s' % (status.system_status, status.instance_status))  # noqa
                            check_successful = True
                            break

                    except Exception as e:
                        self.error_log('Waiting instance ready exception: %s' % str(e))  # noqa
                    sleep(1)

                if not check_successful:
                    msg = "System and instance check were not successful"
                    self.warning_log(msg)
                    raise Exception(msg)
            else:
                self.warning_log("Skipping wait until system and instance check performed")  # noqa

            self.info_log('Starting the selenium node server')

            self.private_ip = instance.private_ip_address
            self.public_dns = instance.public_dns_name
            self.private_dns = instance.private_dns_name
            self.public_ip = instance.ip_address

            # LINUX
            if self.browser_config.get('platform').lower() == "linux":
                command = self.browser_config.get(
                    "selenium_command"
                ).format(**self.browser_config.config)
                self.execute_command(command)

            elif self.browser_config.get('platform').upper() == "windows":

                # TODO this code is out of date
                config = self.browser_config.config.copy()
                config['instance_ip_address'] = instance.ip_address
                command = self.browser_config(
                    "selenium_command"
                ).format(**config)
                process = Popen(
                    command.split(" "),
                    stdout=devnull,
                    stderr=devnull
                )
                self.runner.xvfb_pids.append(process.pid)

            else:

                msg = "The provided platform name is not supported: select either (WINDOWS) or (LINUX)"  # noqa
                self.critical_log(msg)
                raise Exception(msg)

            return True

        except Exception as e:
            self.error_log('Startup exception: %s' % str(e))
            raise

    def tear_down(self):
        """Tear down the instance
        """
        import boto.ec2

        if not self.browser_config.get('terminate'):
            self.warning_log("Skipping terminate")
            return

        self.info_log("Tearing down...")

        ec2 = boto.ec2.connect_to_region(self.browser_config.get("region"))
        ec2.terminate_instances(instance_ids=[self.instance_id])

    def start_video_recording(self, local_video_file_path, video_filename):
        """Start the video recording
        """

        self.runner.info_log("Starting recordscreen...")

        self.local_video_recording_file_path = local_video_file_path
        self.remote_video_recording_file_path = video_filename

        self.execute_command(
            "./start_recording.sh '%s'" % self.remote_video_recording_file_path
        )

    def stop_video_recording(self):
        """Stop the video recording
        """

        self.runner.info_log("Stopping recordscreen...")

        self.execute_command("./stop_recording.sh")
        # self.runner.info_log("output: %s"%output)

        sleep(5)

        self.scp_file_remote_to_local(
            self.remote_video_recording_file_path,
            self.local_video_recording_file_path
        )

    def start_proxy(self):
        """Start the mitmproxy
        """

        self.runner.info_log("Starting proxy...")

        self.proxy_port = self.browser_config.get('proxy_port', 8080)

        self.network_data_path = os.path.join(
            self.runner.runner_dir,
            'network_capture'
        )
        create_dir_if_doesnt_exist(self.network_data_path)

        self.local_proxy_output_path = os.path.join(
            self.network_data_path,
            string_to_filename('%s.data' % self.testname)
        )
        self.local_proxy_log_path = os.path.join(
            self.network_data_path,
            string_to_filename('%s.mitm' % self.testname)
        )

        self.remote_proxy_output_path = string_to_filename(
            '%s.data' % self.testname
        )
        self.remote_proxy_log_path = string_to_filename(
            '%s.mitm' % self.testname
        )

        path_to_mitmproxy = self.browser_config.get(
            "mitmproxy:path",
            'mitmdump'
        )

        filter_ = self.browser_config.get("mitmproxy:filter")
        command = [
            'DISPLAY=:0',
            'nohup',
            path_to_mitmproxy,
            "-p",
            "%s" % str(self.proxy_port),
            "-w",
            "'%s'" % self.remote_proxy_output_path
        ]

        if filter_:
            command.append(filter_)

        command.extend(['>', "'%s'" % self.remote_proxy_log_path, '2>&1'])
        command.append('&')

        self.execute_command(' '.join(command))

    def stop_proxy(self):
        """Stop the mitmproxy
        """

        self.runner.info_log("Stopping proxy...")

        files = [
            (self.remote_proxy_output_path, self.local_proxy_output_path),
            (self.remote_proxy_log_path, self.local_proxy_log_path)
        ]
        for remote_file_path, local_file_path in files:
            self.scp_file_remote_to_local(remote_file_path, local_file_path)

        # kill the proxy
        self.execute_command("fuser -k %s/tcp" % str(self.proxy_port))

    def get_id(self):
        return '%s - %s' % (self.browser_config.browser_id, self.index)

    def debug_log(self, msg):
        self.runner.debug_log("[%s]%s" % (self.get_id(), msg))

    def info_log(self, msg):
        self.runner.info_log("[%s]%s" % (self.get_id(), msg))

    def warning_log(self, msg):
        self.runner.warning_log("[%s]%s" % (self.get_id(), msg))

    def error_log(self, msg):
        self.runner.error_log("[%s]%s" % (self.get_id(), msg))

    def critial_log(self, msg):
        self.runner.critial_log("[%s]%s" % (self.get_id(), msg))



class BaseInstance(object):

    def get_ip(self):
        pass

    def execute_command(self, command):
        pass

    def startup(self):
        pass

    def tear_down(self):
        pass

    def start_proxy(self):
        pass

    def stop_proxy(self):
        pass

    def start_video_recording(self, local_video_file_path, video_filename):
        pass

    def stop_video_recording(self):
        pass

    def debug_log(self, msg):
        print(msg)

    def info_log(self, msg):
        print(msg)

    def warning_log(self, msg):
        print(msg)

    def error_log(self, msg):
        print(msg)

    def critial_log(self, msg):
        print(msg)



import traceback

from brome.runner.localhost_instance import LocalhostInstance
from brome.core.settings import BROME_CONFIG
from brome.runner.base_runner import BaseRunner
from brome.runner.browser_config import BrowserConfig
from brome.model.testbatch import Testbatch
from brome.core.utils import (
    utcnow,
    DbSessionContext
)


class LocalRunner(BaseRunner):
    """The local runner only run browser on localhost
    """

    def __init__(self, *args):
        super(LocalRunner, self).__init__(*args)

    def execute(self):
        """Execute the test batch
        """

        self.browser_config = BrowserConfig(
            runner=self,
            browser_id=BROME_CONFIG['runner_args']['localhost_runner'],
            browsers_config=BROME_CONFIG['browsers_config']
        )

        try:
            self.run()
        except KeyboardInterrupt:
            self.info_log("Test batch interrupted")

        except:
            tb = traceback.format_exc()
            self.error_log("Exception in run of the grid runner: %s" % str(tb))
            raise
        finally:
            self.terminate()

    def run(self):
        """Run the test batch
        """

        self.info_log("The test batch is ready.")

        self.executed_tests = []

        for test in self.tests:
            localhost_instance = LocalhostInstance(
                runner=self,
                browser_config=self.browser_config,
                test_name=test.Test.name
            )

            localhost_instance.startup()

            with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                test_batch = session.query(Testbatch)\
                    .filter(Testbatch.mongo_id == self.test_batch_id).one()
                test_batch.total_executing_tests = 1
                session.save(test_batch, safe=True)

            test_ = test.Test(
                runner=self,
                browser_config=self.browser_config,
                name=test.Test.name,
                test_batch_id=self.test_batch_id,
                localhost_instance=localhost_instance,
                index=1
            )

            test_.execute()
            self.executed_tests.append(test_)

            localhost_instance.tear_down()

    def terminate(self):
        """Terminate the test batch
        """

        self.info_log('The test batch is finished.')

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == self.test_batch_id).one()
            test_batch.ending_timestamp = utcnow()
            test_batch.terminated = True
            session.save(test_batch, safe=True)

        self.print_test_summary(self.executed_tests)



from tempfile import tempdir
from inspect import currentframe
import re
import os

from pdb import set_trace
from selenium.webdriver.common.action_chains import ActionChains
from selenium.common import exceptions
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from brome.core.utils import (
    say,
    utcnow,
    string_to_filename,
    DbSessionContext,
    get_timestamp
)
# from brome.core.bot_diary import BotDiary
from brome.core.settings import BROME_CONFIG
from brome.core.selector import Selector
from brome.core.proxy_element import ProxyElement
from brome.core.proxy_element_list import ProxyElementList
from brome.model.test import Test
from brome.model.testresult import Testresult
from brome.model.testqualityscreenshot import Testqualityscreenshot
from brome.model.testscreenshot import Testscreenshot


class ProxyDriver(object):
    """This class act as a proxy between the driver calls and the selenium driver

    It add functionalities or add value to existing selenium functionalities

    If the proxy driver doesn't have a particular method it
    automatically redirect it to the selenium native driver

    If you need to bypass the proxy driver you can by accessing the
    _driver directly:
        e.g.::

            pdriver._driver.find_element_by_xpath("//div")

    Attributes:
        driver (object): the native selenium driver
        test_instance (object): the test instance binded to this driver
        runner (object): the runner instance binded to this driver
    """

    def __init__(self, driver, test_instance, runner):
        self._driver = driver
        self.test_instance = test_instance
        self.runner = runner

        if BROME_CONFIG['bot_diary']['enable_bot_diary']:
            # TODO rewrite the whole thing
            # self.bot_diary = BotDiary(self)
            self.bot_diary = False
        else:
            self.bot_diary = False

        self.browser_config = self.test_instance._browser_config

        # Use when we run with the remote runner
        # We don't want to embed in this case
        self.embed_disabled = False

        self.no_javascript_error_string = 'No javascript error'

    def __getattr__(self, funcname):
        """Redirect to the native selenium driver when necessary
        """

        return getattr(self._driver, funcname)

    # IS
    def is_present(self, selector):
        """Check if an element is present in the dom or not

        This method won't check if the element is displayed or not
        This method won't wait until the element is visible or present
        This method won't raise any exception if the element is not present

        Returns:
            bool: True if the element is present; False otherwise
        """
        self.debug_log("Is present (%s)" % selector)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_present=False,
            wait_until_visible=False
        )
        if element:
            self.debug_log("is present: True")
            return True
        else:
            self.debug_log("is present: False")
            return False

    def is_visible(self, selector):
        """Check if an element is visible in the dom or not

        This method will check if the element is displayed or not

        This method might (according
        to the config highlight:element_is_visible)
        highlight the element if it is visible

        This method won't wait until the element is visible or present
        This method won't raise any exception if the element is not visible

        Returns:
            bool: True if the element is visible; False otherwise
        """

        self.debug_log("Is visible (%s)" % selector)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_present=False,
            wait_until_visible=False
        )

        if element:
            if element.is_displayed(raise_exception=False):

                element.highlight(
                    style=BROME_CONFIG['highlight']['element_is_visible']
                )

                self.debug_log("is visible (%s): True" % selector)

                return True

        self.debug_log("is visible (%s): False" % selector)

        return False

    # FIND
    def find(self, selector, **kwargs):
        """Find an element with a selector

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            wait_until_present (bool)
            wait_until_visible (bool)
            raise_exception (bool)

        Returns:
            None if no element was found
            proxy_element is an element was found

        Raises:
            this function might raise an exception
                depending on the raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """

        self.debug_log("Finding element with selector: %s" % selector)

        elements = self.find_all(selector, **kwargs)
        if len(elements):
            self.debug_log("find (%s): Element found" % (selector))
            return elements[0]
        else:
            self.debug_log("find (%s): No element found" % (selector))
            return None

    def find_last(self, selector, **kwargs):
        """Return the last element found with a selector

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            wait_until_present (bool)
            wait_until_visible (bool)
            raise_exception (bool)

        Returns:
            None if no element was found
            proxy_element is an element was found

        Raises:
            this function might raise an exception depending
                on the raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """

        self.debug_log("Finding last element with selector: %s" % selector)

        elements = self.find_all(selector, **kwargs)
        if len(elements):
            self.debug_log("find_last (%s): element found" % selector)
            return elements[-1]
        else:
            self.debug_log("find_last (%s): No element found" % selector)
            return None

    def find_all(self, selector, **kwargs):
        """Return all the elements found with a selector

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            wait_until_present (bool) default configurable via
                proxy_driver:wait_until_present_before_find
            wait_until_visible (bool) default configurable via
                proxy_driver:wait_until_visible_before_find
            raise_exception (bool) default configurable via
                proxy_driver:raise_exception

        Returns:
            empty list if no element was found
            proxy_element_list when element are found

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """

        self.debug_log("Finding elements with selector: %s" % selector)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        wait_until_present = kwargs.get(
            'wait_until_present',
            BROME_CONFIG['proxy_driver']['wait_until_present_before_find']
        )

        self.debug_log(
            "effective wait_until_present: %s" % wait_until_present
        )

        wait_until_visible = kwargs.get(
            'wait_until_visible',
            BROME_CONFIG['proxy_driver']['wait_until_visible_before_find']
        )
        self.debug_log(
            "effective wait_until_visible: %s" % wait_until_visible
        )

        _selector = Selector(self, selector)

        found = False
        if wait_until_visible:
            # we don't raise exception here otherwise none visible
            # element will always raise exception
            # TODO find a good way to make it configurable
            found = self.wait_until_visible(selector, raise_exception=False)

        if wait_until_present and not found:
            found = self.wait_until_present(
                selector,
                raise_exception=raise_exception
            )
            if not found:
                self.debug_log("find_all (%s): No element found" % _selector)
                return []

        try:
            elements = getattr(
                self._driver,
                _selector.find_function
            )(_selector.get_selector())
        except exceptions.NoSuchElementException:
            self.debug_log("find_all (%s): No element found" % _selector)
            self.print_javascript_error()
            if raise_exception:
                raise exceptions.NoSuchElementException(_selector)
            else:
                return []

        if type(elements) == list:
            if len(elements):
                self.debug_log("find_all (%s): Element found" % _selector)
                return ProxyElementList(elements, _selector, self)
            else:
                msg = "find_all (%s): No element found" % _selector
                self.debug_log(msg)
                self.print_javascript_error()
                if raise_exception:
                    raise exceptions.NoSuchElementException(msg)
                else:
                    return []
        else:
            self.debug_log("find_all (%s): Element found" % _selector)
            return [ProxyElement(elements, _selector, self)]

    # WAIT
    def wait_until_clickable(self, selector, **kwargs):
        """Wait until an element is clickable

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            timeout (int) second before a timeout exception is raise
            raise_exception (bool) raise an exception or return a bool

        Returns:
            bool True if element is clickable before the timeout,
                False otherwise (raise_exception = False)

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """
        self.info_log("Waiting until clickable (%s)" % selector)

        timeout = kwargs.get(
            'timeout',
            BROME_CONFIG['proxy_driver']['default_timeout']
        )
        self.debug_log("effective timeout: %s" % timeout)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        _selector = Selector(self, selector)

        try:
            WebDriverWait(
                self._driver, timeout
            ).until(
                EC.element_to_be_clickable(
                    (getattr(By, _selector.find_by), _selector.get_selector()))
            )
            self.debug_log(
                "wait_until_clickable (%s): element is clickable" % _selector
            )
            return True
        except exceptions.TimeoutException:
            msg = "wait_until_clickable: element (%s) is still not clickable" % _selector  # noqa
            self.debug_log(msg)
            self.print_javascript_error()
            if raise_exception:
                if self.bot_diary:
                    self.bot_diary.add_auto_entry(
                        "I waited for the clickability of",
                        target=_selector.get_human_readable()
                    )
                raise exceptions.TimeoutException(msg)
            else:
                return False

    def wait_until_present(self, selector, **kwargs):
        """Wait until an element is present

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            timeout (int) second before a timeout exception is raise
            raise_exception (bool) raise an exception or return a bool

        Returns:
            proxy_element: if element is present before the timeout,
                False otherwise (raise_exception = False)

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """
        self.info_log(
            "Waiting until present (%s)" % selector
        )

        timeout = kwargs.get(
            'timeout',
            BROME_CONFIG['proxy_driver']['default_timeout']
        )
        self.debug_log("effective timeout: %s" % timeout)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        _selector = Selector(self, selector)
        try:
            el = WebDriverWait(
                self._driver, timeout
            ).until(
                EC.presence_of_element_located(
                    (getattr(By, _selector.find_by), _selector.get_selector())
                )
            )
            self.debug_log(
                "wait_until_present (%s): element is present" % _selector
            )
            return ProxyElement(el, _selector, self)
        except exceptions.TimeoutException:
            msg = "wait_until_present (%s): element is still not present" % _selector  # noqa
            self.debug_log(msg)
            self.print_javascript_error()
            if raise_exception:
                if self.bot_diary:
                    self.bot_diary.add_auto_entry(
                        "I waited for the presence of",
                        target=_selector.get_human_readable()
                    )
                raise exceptions.TimeoutException(msg)
            else:
                return False

    def wait_until_not_present(self, selector, **kwargs):
        """Wait until an element is not present

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            timeout (int) second before a timeout exception is raise
            raise_exception (bool) raise an exception or return a bool

        Returns:
            bool: True if the element is not present before the timeout,
                False otherwise (raise_exception = False)

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """
        self.info_log("Waiting until not present (%s)" % selector)

        timeout = kwargs.get(
            'timeout',
            BROME_CONFIG['proxy_driver']['default_timeout']
        )
        self.debug_log("effective timeout: %s" % timeout)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        _selector = Selector(self, selector)
        try:
            WebDriverWait(
                self._driver, timeout
            ).until(
                EC.invisibility_of_element_located(
                    (getattr(By, _selector.find_by), _selector.get_selector())
                )
            )
            self.debug_log(
                "wait_until_not_present (%s): element is not present" % _selector  # noqa
            )
            return True
        except exceptions.TimeoutException:
            msg = "wait_until_not_present (%s): element is still present" % _selector  # noqa
            self.debug_log(msg)
            self.print_javascript_error()
            if raise_exception:
                if self.bot_diary:
                    self.bot_diary.add_auto_entry(
                        "I waited for the absence of",
                        target=_selector.get_human_readable()
                    )
                raise exceptions.TimeoutException(msg)
            else:
                return False

    def wait_until_visible(self, selector, **kwargs):
        """Wait until an element is visible

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            timeout (int) second before a timeout exception is raise
            raise_exception (bool) raise an exception or return a bool

        Returns:
            proxy_elment: if the element is visible before the timeout,
                False otherwise (raise_exception = False)

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """

        self.info_log("Waiting until visible (%s)" % selector)

        timeout = kwargs.get(
            'timeout',
            BROME_CONFIG['proxy_driver']['default_timeout']
        )
        self.debug_log("effective timeout: %s" % timeout)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        _selector = Selector(self, selector)
        try:
            el = WebDriverWait(
                self._driver, timeout
            ).until(
                EC.visibility_of_element_located(
                    (getattr(By, _selector.find_by), _selector.get_selector())
                )
            )
            self.debug_log(
                "wait_until_visible (%s): element is visible" % _selector
            )
            return ProxyElement(el, _selector, self)
        except exceptions.TimeoutException:
            msg = "wait_until_visible (%s): element is still not visible" % _selector  # noqa
            self.debug_log(msg)
            self.print_javascript_error()
            if raise_exception:
                if self.bot_diary:
                    self.bot_diary.add_auto_entry(
                        "I waited for the visibility of", target=_selector.get_human_readable()  # noqa
                    )
                raise exceptions.TimeoutException(msg)
            else:
                return False

    def wait_until_not_visible(self, selector, **kwargs):
        """Wait until an element is not visible

        Args:
            selector (str): the selector used to find the element

        Kwargs:
            timeout (int) second before a timeout exception is raise
            raise_exception (bool) raise an exception or return a bool

        Returns:
            bool: True if the element is not visible before the timeout,
                False otherwise (raise_exception = False)

        Raises:
            this function might raise an exception depending on the
                raise_exception kwargs
            or
            the config proxy_driver:raise_exception
        """
        self.info_log("Waiting until not visible (%s)" % selector)

        timeout = kwargs.get(
            'timeout',
            BROME_CONFIG['proxy_driver']['default_timeout']
        )
        self.debug_log("effective timeout: %s" % timeout)

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        self.debug_log("effective raise_exception: %s" % raise_exception)

        _selector = Selector(self, selector)
        try:
            WebDriverWait(
                self._driver, timeout
            ).until(
                EC.invisibility_of_element_located(
                    (getattr(By, _selector.find_by), _selector.get_selector())
                )
            )
            self.debug_log(
                "wait_until_not_visible (%s): element is not visible" % _selector  # noqa
            )
            return True
        except exceptions.TimeoutException:
            msg = "wait_until_not_visible (%s): element is still visible" % _selector  # noqa
            self.debug_log(msg)
            self.print_javascript_error()
            if raise_exception:
                if self.bot_diary:
                    self.bot_diary.add_auto_entry(
                        "I waited for the invisibility of",
                        target=_selector.get_human_readable()
                    )
                raise exceptions.TimeoutException(msg)
            else:
                return False

    # FUNCTION
    def configure_resolution(self):
        # This is not supported on android
        if self.get_platform().lower() == 'android':
            return

        # Maximaze window
        if BROME_CONFIG['browser']['maximize_window']:
            self._driver.maximize_window()
        elif BROME_CONFIG['browser'].get('dont_resize', False):
            return
        else:
            # Window position
            self._driver.set_window_position(
                BROME_CONFIG['browser']['window_x_position'],
                BROME_CONFIG['browser']['window_y_position']
            )

            # Window size
            self._driver.set_window_size(
                BROME_CONFIG['browser']['window_width'],
                BROME_CONFIG['browser']['window_height']
            )

    def get_ip_of_node(self, **kwargs):
        if self.browser_config.location in ['localhost', 'appium']:
            return '127.0.0.1'

        elif self.browser_config.location == 'virtualbox':
            return self.runner.instances[self.browser_config.browser_id][0].get_ip()  # noqa

        # EC2
        elif self.browser_config.location == 'ec2':
            if self.browser_config.get('browserName') == 'dummy':
                return '127.0.0.1'
            else:
                exception_str = ''
                try:
                    self._driver.execute_script("error")
                except exceptions.WebDriverException as e:
                    exception_str = str(e)

                try:
                    return re.search("ip: '([^']*)", exception_str).group(1)
                except AttributeError:
                    msg = "The ip address of the node could not be determined"
                    self.error_log(msg)
                    raise Exception(msg)

    def get_id(
                self,
                join_char='-',
                browser_version_join_char='_',
                platform_join_char='_'):

        return join_char.join([
                    self.get_browser_name(),
                    self.get_browser_version(
                        join_char=browser_version_join_char
                    ),
                    self.get_platform(join_char=platform_join_char)
                ])

    def get_browser_name(self):
        if self.browser_config.get('fakeIdentity'):
            return self.browser_config.get('fakeIdentity').title()
        else:
            return self._driver.capabilities['browserName'].title()

    def get_browser_version(self, join_char='_'):
        return self._driver.capabilities['version'].replace('.', join_char)

    def get_platform(self, join_char='_'):
        return self._driver.capabilities['platform'].replace('.', join_char)

    def get(self, url):
        """Navigate to a specific url

        This specific implementation inject a javascript
            script to intercept the javascript error

        Configurable with the "proxy_driver:intercept_javascript_error" config

        Args:
            url (str): the url to navigate to

        Returns:
            bool
        """

        self._driver.get(url)

        if self.bot_diary:
            self.bot_diary.add_auto_entry(
                "I went on",
                target=url,
                take_screenshot=True
            )

        if BROME_CONFIG['proxy_driver']['intercept_javascript_error']:
            self.init_javascript_error_interception()

        return True

    def inject_js_script(self, script_url):
        """inject a javascript script inside the current page

        Arguments:
            script_url: str (path)

        Returns: None
        """

        self._driver.execute_script("""
            var script = document.createElement("script");

            script.src = "%s"

            document.head.appendChild(script);
        """ % script_url)

    def init_javascript_error_interception(self):
        """Inject javascript code that will gather the javascript raised
        """
        self.debug_log("Initializing javascript error interception")

        self._driver.execute_script("""
            window.jsErrors = [];
            window.onerror = function (errorMessage, url, lineNumber) {
                var message = 'Error: ' + errorMessage;
                window.jsErrors.push(message);
                return false;
            };
        """)

    def print_javascript_error(self):
        """Print to the info log the gathered javascript error

        If no error is found then nothing is printed
        """

        errors = self.get_javascript_error(return_type='list')
        if errors:
            self.info_log("Javascript error:")
            for error in errors:
                self.info_log(error)

    def get_javascript_error(self, return_type='string'):
        """Return the gathered javascript error

        Args:
            return_type: 'string' | 'list'; default: 'string'
        """

        if BROME_CONFIG['proxy_driver']['intercept_javascript_error']:
            js_errors = self._driver.execute_script(
                'return window.jsErrors; window.jsErrors = [];'
            )

            if not js_errors:
                js_errors = []

            if return_type == 'list':
                if len(js_errors):
                    return js_errors
                else:
                    return []
            else:
                if len(js_errors):
                    return os.linesep.join(js_errors)
                else:
                    return self.no_javascript_error_string
        else:
            if return_type == 'list':
                return []
            else:
                return self.no_javascript_error_string

    def pdb(self):
        """Start the python debugger

        Calling pdb won't do anything in a multithread context
        """
        if self.embed_disabled:
            self.warning_log("Pdb is disabled when runned from the grid runner because of the multithreading")  # noqa
            return False

        if BROME_CONFIG['runner']['play_sound_on_pdb']:
            say(BROME_CONFIG['runner']['sound_on_pdb'])

        set_trace()

    def drag_and_drop(self, source_selector, destination_selector, **kwargs):
        """Drag and drop

        Args:
            source_selector: (str)
            destination_selector: (str)

        Kwargs:
            use_javascript_dnd: bool; default:
                config proxy_driver:use_javascript_dnd
        """
        self.info_log(
            "Drag and drop: source (%s); destination (%s)" %
            (source_selector, destination_selector)
        )

        use_javascript_dnd = kwargs.get(
            "use_javascript_dnd",
            "proxy_driver:use_javascript_dnd"
        )

        source_el = self.find(source_selector)
        destination_el = self.find(destination_selector)

        if use_javascript_dnd:
            try:
                dnd_script = [
                    "function simulate(f,c,d,e){var b,a=null;for(b in eventMatchers)if(eventMatchers[b].test(c)){a=b;break}if(!a)return!1;document.createEvent?(b=document.createEvent(a),a=='HTMLEvents'?b.initEvent(c,!0,!0):b.initMouseEvent(c,!0,!0,document.defaultView,0,d,e,d,e,!1,!1,!1,!1,0,null),f.dispatchEvent(b)):(a=document.createEventObject(),a.detail=0,a.screenX=d,a.screenY=e,a.clientX=d,a.clientY=e,a.ctrlKey=!1,a.altKey=!1,a.shiftKey=!1,a.metaKey=!1,a.button=1,f.fireEvent('on'+c,a));return!0} var eventMatchers={HTMLEvents:/^(?:load|unload|abort|error|select|change|submit|reset|focus|blur|resize|scroll)$/,MouseEvents:/^(?:click|dblclick|mouse(?:down|up|over|move|out))$/};",  # noqa
                    "var source = arguments[0],destination = arguments[1];",
                    "simulate(source, 'mousedown', 0, 0);",
                    "simulate(source, 'mousemove', destination.offsetLeft, destination.offsetTop);",  # noqa
                    "simulate(source, 'mouseup', destination.offsetLeft, destination.offsetTop);"  # noqa
                ]
                self._driver.execute_script(
                    '\n'.join(dnd_script),
                    source_el._element,
                    destination_el._element
                )

            except Exception as e:
                self.error_log(u'drag_and_drop exception: %s' % str(e))
                raise
        else:
            try:
                ActionChains(self._driver).drag_and_drop(
                    source_el,
                    destination_el
                ).perform()
            except Exception as e:
                self.error_log(u'drag_and_drop exception: %s' % str(e))
                raise

    def embed(self, title=''):
        """Start an IPython embed

        Calling embed won't do anything in a multithread context

        The stack_depth will be found automatically
        """
        if self.embed_disabled:
            self.warning_log("Embed is disabled when runned from the grid runner because of the multithreading")  # noqa
            return False

        from IPython.terminal.embed import InteractiveShellEmbed

        if BROME_CONFIG['runner']['play_sound_on_ipython_embed']:
            say(BROME_CONFIG['runner']['sound_on_ipython_embed'])

        ipshell = InteractiveShellEmbed(banner1=title)

        frame = currentframe()
        stack_depth = 1
        for i in range(5):
            frame = frame.f_back
            stack_depth += 1
            if frame.f_code.co_filename not in __file__:
                break

        msg = 'Stopped at %s and line %s;' % \
            (frame.f_code.co_filename, frame.f_lineno)

        ipshell(msg, stack_depth=stack_depth)

    def take_node_screenshot(self, element, screenshot_path):
        from PIL import Image
        """Take a screenshot of a node

        Args:
            element (object): the proxy_element
            screenshot_path (str): the path where the screenshot will be saved
        """

        temp_path = os.path.join(tempdir, screenshot_path)

        el_x = int(element.location['x'])
        el_y = int(element.location['y'])
        el_height = int(element.size['height'])
        el_width = int(element.size['width'])

        if el_height == 0 or el_width == 0:
            self.debug_log("take_node_screenshot cannot be taken because element width or height equal zero")  # noqa
            return False

        bounding_box = (
            el_x,
            el_y,
            (el_x + el_width),
            (el_y + el_height)
        )

        self._driver.save_screenshot(temp_path)

        base_image = Image.open(temp_path)

        cropped_image = base_image.crop(bounding_box)

        base_image = base_image.resize(cropped_image.size)

        base_image.paste(cropped_image, (0, 0))

        base_image.save(screenshot_path)

        """
        except Exception as e:
            tb = traceback.format_exc()
            print unicode(tb)
            embed()
        """

    def take_screenshot(self, screenshot_name=None, screenshot_path=None):
        """Take a screenshot

        Use the screenshot_name args when you want to take a
            screenshot for reference

        If the `runner:cache_screenshot` config is set to True then
            screenshot sharing all the same name will be saved only once

        The screenshot_path args is exclusively used by the
            proxy_driver:create_test_result function

        Args:
            screenshot_name (str) the name of the screenshot
            screenshot_path (str) the path of the screenshot
        """
        self.info_log("Taking a screenshot...")

        save_to_db = False
        if screenshot_path:
            self._driver.save_screenshot(screenshot_path)
            self.debug_log("Screenshot taken (%s)" % screenshot_path)

        elif screenshot_name:
            take_screenshot = True
            if hasattr(self.runner, "screenshot_cache"):
                if self.runner.screenshot_cache.get(screenshot_name):
                    self.debug_log(
                        "screenshot(%s) found in cache" % screenshot_name
                    )
                    take_screenshot = False

            if take_screenshot:
                if self.test_instance._runner_dir:
                    _screenshot_name = '%s.png' % \
                        string_to_filename(screenshot_name)

                    relative_path = os.path.join(
                            self.test_instance._screenshot_relative_dir,
                            _screenshot_name
                    )

                    full_path = os.path.join(
                            self.test_instance._screenshot_dir,
                            _screenshot_name
                    )

                    self._driver.save_screenshot(
                        full_path
                    )
                    self.debug_log("Screenshot taken (%s)" % full_path)
                    save_to_db = True
        else:
            if self.test_instance._runner_dir:
                screenshot_name = get_timestamp()
                _screenshot_name = '%s.png' % screenshot_name

                relative_path = os.path.join(
                        self.test_instance._screenshot_relative_dir,
                        _screenshot_name
                )

                full_path = os.path.join(
                        self.test_instance._screenshot_dir,
                        _screenshot_name
                )

                self._driver.save_screenshot(
                    full_path
                )
                self.debug_log("Screenshot taken (%s)" % full_path)
                save_to_db = True

        if save_to_db:
            with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                capabilities = {
                    'browserName': self.capabilities['browserName'],
                    'platform': self.capabilities['platform'],
                    'version': self.capabilities['version']
                }
                screenshot = Testscreenshot()
                screenshot.browser_capabilities = capabilities
                screenshot.browser_id = self.get_id()
                # TODO support s3
                screenshot.location = 'local_file_system'
                screenshot.root_path = self.test_instance._runner.root_test_result_dir  # noqa
                screenshot.file_path = relative_path
                screenshot.extra_data = {}
                screenshot.title = screenshot_name
                screenshot.test_instance_id = self.test_instance._test_instance_id  # noqa
                screenshot.test_batch_id = self.test_instance._test_batch_id  # noqa

                session.save(screenshot, safe=True)

    def take_quality_screenshot(self, screenshot_name):
        """Take a quality screenshot

        Use the screenshot_name args when you want to take a
            screenshot for reference

        Args:
            screenshot_name (str) the name of the screenshot
        """
        self.info_log("Taking a quality screenshot...")

        if self.test_instance._runner_dir:
            _screenshot_name = '%s.png' % string_to_filename(screenshot_name)
            relative_path = os.path.join(
                    self.test_instance._quality_screenshot_relative_dir,
                    _screenshot_name
                )

            full_path = os.path.join(
                    self.test_instance._quality_screenshot_dir,
                    _screenshot_name
                )
            self._driver.save_screenshot(
                full_path
            )

            with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                capabilities = {
                    'browserName': self.capabilities['browserName'],
                    'platform': self.capabilities['platform'],
                    'version': self.capabilities['version']
                }
                quality_screenshot = Testqualityscreenshot()
                quality_screenshot.timestamp = utcnow()
                quality_screenshot.browser_capabilities = capabilities
                quality_screenshot.browser_id = self.get_id()
                quality_screenshot.file_path = relative_path
                # TODO support s3
                quality_screenshot.location = 'local_file_system'
                quality_screenshot.root_path = self.test_instance._runner.root_test_result_dir  # noqa
                quality_screenshot.extra_data = {}
                quality_screenshot.title = screenshot_name
                quality_screenshot.test_instance_id = self.test_instance._test_instance_id  # noqa
                quality_screenshot.test_batch_id = self.test_instance._test_batch_id  # noqa

                session.save(quality_screenshot, safe=True)

            self.debug_log("Quality screenshot taken (%s)" % full_path)

    # ASSERT
    def assert_present(self, selector, testid=None, **kwargs):
        """Assert that the element is present in the dom

        Args:
            selector (str): the selector used to find the element
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_present (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert present selector(%s) testid(%s)" % (selector, testid)
        )

        wait_until_present = kwargs.get(
            'wait_until_present',
            BROME_CONFIG['proxy_driver']['wait_until_present_before_assert_present']  # noqa
        )
        self.debug_log(
            "effective wait_until_present: %s" % wait_until_present
        )

        if wait_until_present:
            element = self.wait_until_present(selector, raise_exception=False)
        else:
            element = self.is_present(selector)

        if element:
            if testid is not None:
                self.create_test_result(testid, True)

            return True
        else:
            if testid is not None:
                self.create_test_result(testid, False)

            return False

    def assert_not_present(self, selector, testid=None, **kwargs):
        """Assert that the element is not present in the dom

        Args:
            selector (str): the selector used to find the element
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_not_present (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert not present selector(%s) testid(%s)" %
            (selector, testid)
        )

        wait_until_not_present = kwargs.get(
            'wait_until_not_present',
            BROME_CONFIG['proxy_driver']['wait_until_not_present_before_assert_not_present']  # noqa
        )
        self.debug_log(
            "effective wait_until_not_present: %s" % wait_until_not_present
        )

        if wait_until_not_present:
            ret = self.wait_until_not_present(selector, raise_exception=False)
        else:
            ret = not self.is_present(selector)

        if ret:
            if testid is not None:
                self.create_test_result(testid, True)

            return True
        else:
            if testid is not None:
                self.create_test_result(testid, False)

            return False

    def assert_visible(self, selector, testid=None, **kwargs):
        """Assert that the element is visible in the dom

        Args:
            selector (str): the selector used to find the element
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_visible (bool)
            highlight (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert visible selector(%s) testid(%s)" % (selector, testid)
        )

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_on_assertion_success']
        )
        self.debug_log("effective highlight: %s" % highlight)

        wait_until_visible = kwargs.get(
            'wait_until_visible',
            BROME_CONFIG['proxy_driver']['wait_until_visible_before_assert_visible']  # noqa
        )
        self.debug_log("effective wait_until_visible: %s" % wait_until_visible)

        if wait_until_visible:
            self.wait_until_visible(selector, raise_exception=False)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_visible=False,
            wait_until_present=False
        )
        if element and element.is_displayed(raise_exception=False):
            if highlight:
                element.highlight(
                    style=BROME_CONFIG['highlight']['style_on_assertion_success']  # noqa
                )
            if testid is not None:
                self.create_test_result(testid, True)

            return True
        else:
            if testid is not None:
                self.create_test_result(testid, False)

            return False

    def assert_not_visible(self, selector, testid=None, **kwargs):
        """Assert that the element is not visible in the dom

        Args:
            selector (str): the selector used to find the element
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_not_visible (bool)
            highlight (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert not visible selector(%s) testid(%s)" % (selector, testid)
        )

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_on_assertion_failure']
        )
        self.debug_log("effective highlight: %s" % highlight)

        wait_until_not_visible = kwargs.get(
            'wait_until_not_visible',
            BROME_CONFIG['proxy_driver']['wait_until_not_visible_before_assert_not_visible']  # noqa
        )
        self.debug_log(
            "effective wait_until_not_visible: %s" % wait_until_not_visible
        )

        if wait_until_not_visible:
            self.wait_until_not_visible(selector, raise_exception=False)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_visible=False,
            wait_until_present=False
        )
        if element and element.is_displayed(raise_exception=False):
            data = self.execute_script(
                "return arguments[0].getBoundingClientRect();",
                element._element
            )

            if highlight:
                element.highlight(
                    style=BROME_CONFIG['highlight']['style_on_assertion_failure']  # noqa
                )
            if testid is not None:
                self.create_test_result(testid, False, extra_data={
                    'bounding_client_rect': data,
                    'video_x_offset': self.browser_config.get('video_x_offset', 0),  # noqa
                    'video_y_offset': self.browser_config.get('video_y_offset', 0)  # noqa
                })

            return False
        else:
            if testid is not None:
                self.create_test_result(testid, True)

            return True

    def assert_text_equal(self, selector, value, testid=None, **kwargs):
        """Assert that the element's text is equal to the provided value

        Args:
            selector (str): the selector used to find the element
            value (str): the value that will be compare
                with the element.text value
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_visible (bool)
            highlight (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert text equal selector(%s) testid(%s)" % (selector, testid)
        )

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_on_assertion_success']
        )
        self.debug_log("effective highlight: %s" % highlight)

        wait_until_visible = kwargs.get(
            'wait_until_visible',
            BROME_CONFIG['proxy_driver']['wait_until_visible_before_assert_visible']  # noqa
        )
        self.debug_log("effective wait_until_visible: %s" % wait_until_visible)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_visible=wait_until_visible
        )
        if element:
            if element.text == value:
                if highlight:
                    element.highlight(
                        highlight=BROME_CONFIG['highlight']['style_on_assertion_success']  # noqa
                    )
                if testid is not None:
                    self.create_test_result(testid, True)

                return True
            else:
                if highlight:
                    element.highlight(
                        style=BROME_CONFIG['highlight']['style_on_assertion_failure']  # noqa
                    )
                if testid is not None:
                    self.create_test_result(testid, False)

                return False
        else:
            if testid is not None:
                self.create_test_result(testid, False)

            return False

    def assert_text_not_equal(self, selector, value, testid=None, **kwargs):
        """Assert that the element's text is not equal to the provided value

        Args:
            selector (str): the selector used to find the element
            value (str): the value that will be compare with
                the element.text value
            test_id (str): the test_id or a str

        Kwargs:
            wait_until_visible (bool)
            highlight (bool)

        Returns:
            bool: True is the assertion succeed; False otherwise.
        """
        self.info_log(
            "Assert text not equal selector(%s) testid(%s)" %
            (selector, testid)
        )

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_on_assertion_success']
        )
        self.debug_log("effective highlight: %s" % highlight)

        wait_until_visible = kwargs.get(
            'wait_until_visible',
            BROME_CONFIG['proxy_driver']['wait_until_visible_before_assert_visible']  # noqa
        )
        self.debug_log("effective wait_until_visible: %s" % wait_until_visible)

        element = self.find(
            selector,
            raise_exception=False,
            wait_until_visible=wait_until_visible
        )
        if element:
            if element.text != value:
                if highlight:
                    element.highlight(
                        style=BROME_CONFIG['highlight']['style_on_assertion_success']  # noqa
                    )
                if testid is not None:
                    self.create_test_result(testid, True)

                return True
            else:
                if highlight:
                    element.highlight(
                        BROME_CONFIG['highlight']['style_on_assertion_failure']  # noqa
                    )
                if testid is not None:
                    self.create_test_result(testid, False)

                return False
        else:
            if testid is not None:
                self.create_test_result(testid, False)

            return False

    def create_test_result(self, testid, result, **kwargs):
        """Create a test result entry in the persistence layer

        Args:
            testid (str)
            result (bool)

        Keyword Args:
            extra_data (dict): the extra data that will be
                saved with the test result

        Returns:
            None
        """

        embed = True
        videocapture_path = self.test_instance._video_capture_file_relative_path  # noqa
        screenshot_relative_path = ''
        extra_data = {}

        # JAVASCRIPT ERROR
        if not result:
            extra_data['javascript_error'] = self.get_javascript_error()

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test = None
            if testid in BROME_CONFIG['test_dict']:
                query = session.query(Test).filter(Test.test_id == testid)
                if query.count():
                    test = query.one()

                    test_config = BROME_CONFIG['test_dict'][testid]
                    if type(test_config) == dict:
                        if 'embed' in test_config:
                            embed = test_config['embed']
                            test_name = test_config['name']
                    else:
                        test_name = test_config

                    embed_title = '[%s] %s' % (testid, test_name)

            if not test:
                test_name = testid
                embed_title = test_name

            if result:
                # SCREENSHOT
                if BROME_CONFIG['proxy_driver']['take_screenshot_on_assertion_success']:  # noqa
                    if self.test_instance._runner_dir:
                        screenshot_name = 'succeed_%s_%s_%s.png' % (
                            string_to_filename(testid),
                            get_timestamp(),
                            self.get_id(join_char='_')
                        )
                        screenshot_path = os.path.join(
                            self.test_instance._assertion_screenshot_dir,
                            screenshot_name
                        )
                        screenshot_relative_path = os.path.join(
                            self.test_instance._assertion_screenshot_relative_dir,  # noqa
                            screenshot_name
                        )
                        self.take_screenshot(screenshot_path=screenshot_path)

                # SOUND NOTIFICATION
                if BROME_CONFIG['runner']['play_sound_on_assertion_success']:  # noqa
                    say(
                        BROME_CONFIG['runner']['sound_on_assertion_success']
                        .format(testid=testid)
                    )

                # EMBED
                if BROME_CONFIG['runner']['embed_on_assertion_success'] and embed:  # noqa
                    self.embed(title=embed_title)
            else:
                # SCREENSHOT
                if BROME_CONFIG['proxy_driver']['take_screenshot_on_assertion_failure']:  # noqa
                    if self.test_instance._runner_dir:
                        screenshot_name = 'failed_%s_%s_%s.png' % (
                            string_to_filename(testid),
                            get_timestamp(),
                            self.get_id(join_char='_')
                        )
                        screenshot_path = os.path.join(
                            self.test_instance._assertion_screenshot_dir,
                            screenshot_name
                        )
                        screenshot_relative_path = os.path.join(
                            self.test_instance._assertion_screenshot_relative_dir,  # noqa
                            screenshot_name
                        )
                        self.take_screenshot(screenshot_path=screenshot_path)

                # SOUND NOTIFICATION
                if BROME_CONFIG['runner']['play_sound_on_assertion_failure']:  # noqa
                    say(
                        BROME_CONFIG['runner']['sound_on_assertion_failure']
                        .format(testid=testid)
                    )

                # EMBED
                if BROME_CONFIG['runner']['embed_on_assertion_failure'] and embed:  # noqa
                    self.embed(title=embed_title)

            capabilities = {
                'browserName': self.capabilities['browserName'],
                'platform': self.capabilities['platform'],
                'version': self.capabilities['version']
            }
            test_result = Testresult()
            test_result.result = result
            test_result.timestamp = utcnow()
            test_result.browser_capabilities = capabilities
            test_result.browser_id = self.get_id()
            test_result.root_path = self.test_instance._runner.root_test_result_dir  # noqa
            test_result.screenshot_path = screenshot_relative_path
            test_result.video_capture_path = videocapture_path
            test_result.extra_data = extra_data
            test_result.title = test_name
            if test:
                test_result.test_id = test.get_uid()
                test_result.testid = test.test_id
            else:
                test_result.testid = test_name

            test_result.test_instance_id = self.test_instance._test_instance_id
            test_result.test_batch_id = self.runner.test_batch_id

            session.save(test_result, safe=True)

    # LOG
    def debug_log(self, msg):
        self.test_instance.debug_log(msg)

    def info_log(self, msg):
        self.test_instance.info_log(msg)

    def warning_log(self, msg):
        self.test_instance.warning_log(msg)

    def error_log(self, msg):
        self.test_instance.error_log(msg)

    def critical_log(self, msg):
        self.test_instance.critical_log(msg)



from mongoalchemy.fields import (
    StringField
)

from brome.model.basemodel import BaseModel
from brome.core import exceptions


class Test(BaseModel):

    test_id = StringField()
    name = StringField()

    def __repr__(self):
        try:
            return "Test <test_id: {self.test_id}><name: {self.name}>".format(
                self=self
            )
        except AttributeError:
            return "Test uninitialized"

    async def sanitize_data(self, context):
        return []

    async def serialize(self, context):
        data = {}
        data['uid'] = self.get_uid()
        data['test_id'] = self.test_id
        data['name'] = self.name
        return data

    async def method_autorized(self, context):
        method = context.get('method')
        if method in ['create', 'update', 'delete']:
            return False
        else:
            return True

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')

        is_new = self.is_new()

        # TEST ID
        test_id = data.get('test_id')
        if test_id is not None:
            self.test_id = test_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_id')

        # NAME
        name = data.get('name')
        if name:
            self.name = name
        else:
            if is_new:
                raise exceptions.MissingModelValueException('name')

        db_session.save(self, safe=True)



import logging
import traceback
from time import sleep
import os
import pickle
import urllib

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common import proxy

from IPython import embed
from brome.core.settings import BROME_CONFIG
from brome.core.stateful import Stateful
from brome.core.proxy_driver import ProxyDriver
from brome.core.configurator import (
    test_config_to_dict,
    parse_brome_config_from_browser_config
)
from brome.model.testinstance import Testinstance
from brome.model.testbatch import Testbatch
from brome.model.testresult import Testresult
from brome.model.testcrash import Testcrash
from brome.model.test import Test
from brome.core import exceptions
from brome.core.utils import (
    DbSessionContext,
    utcnow,
    say,
    create_dir_if_doesnt_exist,
    string_to_filename
)


class BaseTest(object):
    """Base test class used to init the driver, load and save state, start and
        stop video recording

    Attributes:
        runner (object): the runner that created the BaseTest instance
        name (str): the name of the test
        index (int): the index of the test
        browser_config (object): the browser config; used to init the driver
        test_batch_id (int): the test batch id
    """

    def __init__(self, runner, name, index, browser_config, test_batch_id, **kwargs):  # noqa
        self._runner = runner
        self._name = name
        self._index = index
        self._browser_config = browser_config
        self._test_batch_id = test_batch_id
        self._localhost_instance = kwargs.get('localhost_instance')
        self._log_file_path = ''


        browser_config = parse_brome_config_from_browser_config(self._browser_config.config)
        for key in iter(BROME_CONFIG):
            if key in browser_config:
                BROME_CONFIG[key].update(browser_config[key])

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == test_batch_id)\
                .one()
            self._test_batch_friendly_name = test_batch.friendly_name

        self._crash_error = False

        # TEST BATCH DIRECTORY
        if self._runner.runner_dir:
            self._runner_dir = self._runner.runner_dir
        else:
            self._runner_dir = ''

        # NETWORK CAPTURE
        if self._browser_config.get('enable_proxy'):
            self._network_capture_file_relative_path = os.path.join(
                self._runner.relative_runner_dir,
                'network_capture',
                string_to_filename('%s.data' % self._name)
            )
        else:
            self._network_capture_file_relative_path = ''

        # LOGGING
        self.configure_logger()

        # DRIVER
        self.pdriver = self.init_driver()

        # ASSIGN THE TEST NAME TO THE INSTANCE
        if self._browser_config.location in ['ec2', 'virtualbox']:
            self._runner.resolve_instance_by_ip(
                self.pdriver.get_ip_of_node()
            ).testname = self._name

        # TEST RESULT DIRECTORY
        self.configure_test_result_dir()

        # DRIVER RESOLUTION
        self.pdriver.configure_resolution()

        # VIDEO RECORDING
        self.start_video_recording()
        starting_timestamp = utcnow()

        # TEST KWARGS
        self._test_config = test_config_to_dict(
            BROME_CONFIG['runner'].get('test_config')
        )

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa

            extra_data = {}
            if self._browser_config.location in 'ec2':
                self._private_ip = self.pdriver.get_ip_of_node()
                extra_data['instance_private_ip'] = self._runner.instances_ip[self._private_ip].private_ip  # noqa
                extra_data['instance_public_ip'] = self._runner.instances_ip[self._private_ip].public_ip  # noqa
                extra_data['instance_public_dns'] = self._runner.instances_ip[self._private_ip].public_dns  # noqa
                extra_data['instance_private_dns'] = self._runner.instances_ip[self._private_ip].private_dns  # noqa
            elif self._browser_config.location in 'virtualbox':
                self._private_ip = self._browser_config.get('ip')
                extra_data['instance_private_ip'] = self._private_ip

            capabilities = {
                'browserName': self.pdriver.capabilities['browserName'],
                'platform': self.pdriver.capabilities['platform'],
                'version': self.pdriver.capabilities['version']
            }
            test_instance = Testinstance()
            test_instance.name = self._name
            test_instance.starting_timestamp = starting_timestamp
            test_instance.browser_capabilities = capabilities
            test_instance.browser_id = self.pdriver.get_id()
            test_instance.test_batch_id = self._test_batch_id
            test_instance.extra_data = extra_data
            if self._runner.root_test_result_dir:
                test_instance.log_file_path = self._relative_log_file_path
                test_instance.root_path = self._runner.root_test_result_dir
                test_instance.network_capture_path = self._network_capture_file_relative_path  # noqa
                test_instance.video_capture_path = self._video_capture_file_relative_path  # noqa
            else:
                test_instance.log_file_path = ''
                test_instance.root_path = ''
                test_instance.network_capture_path = ''
                test_instance.video_capture_path = ''

            session.save(test_instance, safe=True)

            self._test_instance_id = test_instance.get_uid()

        # START PROXY
        if self._browser_config.location == 'ec2':
            instance = self._runner.resolve_instance_by_ip(self._private_ip)
            if instance.browser_config.get('enable_proxy'):
                instance.start_proxy()

        # FEATURE DETECTION
        self.detect_feature()

    def detect_feature(self):
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == self._test_batch_id).one()

            # SESSION VIDEO CAPTURE
            if self._browser_config.get('record_session'):
                test_batch.feature_session_video_capture = True

            # NETWORK CAPTURE
            if self._browser_config.get('enable_proxy'):
                test_batch.feature_network_capture = True

            # SCREENSHOT
            if self._runner_dir:
                test_batch.feature_style_quality = True
                test_batch.feature_screenshots = True

            # BOT DIARY
            if BROME_CONFIG['bot_diary']['enable_bot_diary']:
                test_batch.feature_bot_diaries = False

            # TODO VNC

            session.save(test_batch, safe=True)

    def start_video_recording(self):
        self._video_capture_file_relative_path = ''

        # TODO support localhost and more
        if self._browser_config.get('record_session') and \
                self._browser_config.location in ['ec2', 'virtualbox']:
            self.info_log("Starting screen capture...")

            video_capture_file = string_to_filename(
                '%s.mp4' % (self._name.replace(' ', '_'))
            )

            self._video_capture_file_path = os.path.join(
                self._video_recording_dir,
                video_capture_file
            )

            self._video_capture_file_relative_path = os.path.join(
                self._video_recording_relative_dir,
                video_capture_file
            )

            private_ip = self.pdriver.get_ip_of_node()
            instance = self._runner.resolve_instance_by_ip(private_ip)
            instance.start_video_recording(
                self._video_capture_file_path,
                video_capture_file
            )

    def stop_video_recording(self):
        if self._browser_config.get('record_session'):
            self.info_log("Stopping the screen capture...")

            instance = self._runner.resolve_instance_by_ip(self._private_ip)
            instance.stop_video_recording()

    def init_driver(self, retry=30):
        """Init driver will instanciate a webdriver according to the browser config

        First a webdriver is instanciate
            according to the provided browser config
        then the webdriver is wrapped in the brome proxy_driver

        Args:
            retry (int): the number of time the init driver will retry to
                init the driver if it fail
                            default: 10

        Returns:
            proxy_driver instance

        Raises:
            Exception InitDriverException
            Exception InvalidBrowserName

        """

        # LOCAL
        if self._browser_config.location == 'localhost':

            def get_proxy():
                mitm_proxy = "localhost:%s" % self._localhost_instance.proxy_port  # noqa

                return proxy.Proxy({
                    'proxyType': proxy.ProxyType.MANUAL,
                    'httpProxy': mitm_proxy,
                    'sslProxy': mitm_proxy
                })

            # CHROME
            if self._browser_config.get('browserName').lower() == 'chrome':
                chrome_options = webdriver.ChromeOptions()

                if self._browser_config.get('mobile_emulation'):
                    chrome_options.add_experimental_option(
                        "mobileEmulation",
                        self._browser_config.get('mobile_emulation')
                    )
                elif self._browser_config.get('enable_proxy'):
                    chrome_options.add_argument(
                        "--proxy-server={0}".format(get_proxy())
                    )

                driver = webdriver.Chrome(chrome_options=chrome_options)

            elif self._browser_config.get('browserName').lower() \
                    in ['firefox', 'phantomjs'] \
                    and self._browser_config.get('enable_proxy'):

                # NOTE http://www.seleniumhq.org/docs/04_webdriver_advanced.jsp#using-a-proxy  # noqa
                # FIREFOX
                if self._browser_config.get('browserName').lower() \
                        == 'firefox':
                    profile = webdriver.FirefoxProfile()
                    profile.set_proxy(proxy=get_proxy())
                    driver = webdriver.Firefox(firefox_profile=profile)

                # PHANTOMJS
                elif self._browser_config.get('browserName').lower() \
                        == 'phantomjs':

                    # TODO investigate why this doesnt work in phantomjs
                    """
                    #WAY 1
                    #http://stackoverflow.com/questions/14699718/how-do-i-set-a-proxy-for-phantomjs-ghostdriver-in-python-webdriver
                    service_args = [
                    '--proxy=127.0.0.1:%s'%self._localhost_instance.proxy_port,
                    '--proxy-type=http',
                    ]
                    driver = webdriver.PhantomJS(service_args=service_args)

                    #WAY 2
                    desired_cap = self._browser_config.config
                    proxy.add_to_capabilities(desired_cap)
                    driver = webdriver.PhantomJS(desired_capabilities = desired_cap) # noqa
                    """
                    raise NotImplemented()

                # NOTE: AFAIK Safari doesn't support proxy
                # https://code.google.com/p/selenium/issues/detail?id=5051
            else:
                try:
                    driver = getattr(
                        webdriver,
                        self._browser_config.get('browserName')
                    )()
                except AttributeError:
                    raise exceptions.InvalidBrowserName(
                        "The browserName('%s') is invalid" %
                        self._browser_config.get('browserName')
                    )

        # APPIUM
        elif self._browser_config.location == 'appium':
            driver = webdriver.Remote(
                command_executor='http://127.0.0.1:4723/wd/hub',
                desired_capabilities=self._browser_config.config
            )

        # SAUCELABS
        elif self._browser_config.location == 'saucelabs':
            driver = webdriver.Remote(
                command_executor='http://%s:%s@ondemand.saucelabs.com:80/wd/hub' % (  # noqa
                    BROME_CONFIG['saucelabs']['username'],
                    BROME_CONFIG['saucelabs']['key']
                ),
                desired_capabilities=self._browser_config.config
            )

        # BROWSERSTACK
        elif self._browser_config.location == 'browserstack':
            driver = webdriver.Remote(
                command_executor='http://%s:%s@hub.browserstack.com:80/wd/hub' % (  # noqa
                    BROME_CONFIG['browserstack']['username'],
                    BROME_CONFIG['browserstack']['key']
                ),
                desired_capabilities=self._browser_config.config
            )

        # REMOTE
        elif self._browser_config.location in ['virtualbox', 'ec2']:
            config = self._browser_config.config

            desired_cap = {}
            desired_cap['browserName'] = config.get('browserName')
            desired_cap['platform'] = config.get('platform')
            desired_cap['javascriptEnabled'] = True

            if config.get('enable_proxy'):
                mitm_proxy = "localhost:%s" % config.get('proxy_port', 8080)

                _proxy = proxy.Proxy({
                    'proxyType': proxy.ProxyType.MANUAL,
                    'httpProxy': mitm_proxy,
                    'sslProxy': mitm_proxy
                })

            if desired_cap['browserName'].lower() == "chrome":
                chrome_options = Options()
                chrome_options.add_argument("--test-type")
                chrome_options.add_argument("--disable-application-cache")

                if config.get('enable_proxy'):
                    chrome_options.add_argument(
                        '--proxy-server=%s' % mitm_proxy
                    )
                if config.get('mobile_emulation'):
                    chrome_options.add_experimental_option(
                        "mobileEmulation",
                        config.get('mobile_emulation')
                    )


                desired_cap = chrome_options.to_capabilities()
                desired_cap['platform'] = config.get('platform')

            try:
                command_executor = "http://%s:%s/wd/hub" % (  # noqa
                    BROME_CONFIG['grid_runner']['selenium_server_ip'],
                    BROME_CONFIG['grid_runner']['selenium_server_port']
                )

                if desired_cap['browserName'].lower() == "firefox" \
                        and config.get('enable_proxy'):
                    profile = webdriver.FirefoxProfile()
                    profile.set_proxy(proxy=_proxy)
                    driver = webdriver.Remote(
                            browser_profile=profile,
                            desired_capabilities=desired_cap,
                            command_executor=command_executor)
                else:
                    driver = webdriver.Remote(
                            command_executor=command_executor,
                            desired_capabilities=desired_cap
                    )

                self.info_log('Got a session')

            except Exception as e:
                if str(e).find("Error forwarding the new session") != -1:
                    if retry:

                        self.info_log("Waiting 5 sec because the pool doesn't contain the needed browser.")  # noqa

                        sleep(5)

                        return self.init_driver(retry=(retry - 1))
                    else:
                        raise exceptions.InitDriverException("Cannot get the driver")  # noqa

        # Wrap the driver in the brome proxy driver and return it
        return ProxyDriver(
            driver=driver,
            test_instance=self,
            runner=self._runner
        )

    def delete_state(self, state_pickle=None):
        """This will delete the pickle that hold the saved test state

        Args:
            state_pickle (path): the path of the pickle that will be delete
        """
        if not state_pickle:
            state_pickle = self.get_state_pickle_path()

        if os.path.isfile(state_pickle):
            os.remove(state_pickle)
            self.info_log("State deleted: %s" % state_pickle)

    def get_state_pickle_path(self):
        """Return the state's pickle path

        The project:url config need to be set in order to
            extract the server name. Each server will have it own state
        The test name is used to name the state.

        E.g.: the test named "test 1" that run with the project:url
            ("http://www.example.com") with have a state named:
        tests/states/test_1_example.com.pkl

        Returns: str (path)
        """

        # Extract the server name
        server = urllib.parse.urlparse(
            BROME_CONFIG['project']['url']
        ).netloc

        if BROME_CONFIG['project']['test_batch_result_path']:
            states_dir = os.path.join(
                BROME_CONFIG['project']['absolute_path'],
                BROME_CONFIG['project']['script_folder_name'],
                "states"
            )
            create_dir_if_doesnt_exist(states_dir)

            # TODO should be configurable
            state_pickle_path = os.path.join(
                states_dir,
                string_to_filename(
                    '%s_%s.pkl' %
                    (
                        self._name.replace(' ', '_'),
                        server
                    )
                )
            )

            return state_pickle_path
        else:
            return False

    def save_state(self):
        """Save the state in a pickle

        First remove all the instance variables that are private (_*).
        Then remove the pdriver instance variable.
        Finally cleanup the state using the Stateful.cleanup_state function.
        The cleanup is recursive.

        The test state will be saved to the path returned by the
            get_state_pickle_path.

        Returns: None
        """
        self.info_log("Saving state...")

        state = {}

        # Remove all the instance variables that are private
        state = {key: value for (key, value) in iter(self.__dict__.items()) if key[0] != '_'}  # noqa

        # Remove the pdriver
        del state['pdriver']

        # Cleanup the state
        effective_state = Stateful.cleanup_state(state)

        # Save the state
        state_pickle_path = self.get_state_pickle_path()
        if state_pickle_path:
            with open(state_pickle_path, 'wb') as fd:
                pickle.dump(effective_state, fd)

            self.info_log("State saved: %s" % state_pickle_path)

    def load_state(self):
        """Load the state pickle into the instance object if a state is found

        The state will be loaded in the basetest instance if a state exist.
        The pdriver will be set on each instance that inherited from Stateful
        These instance can be in a dict or a list. It is not recursive tho.

        If the instance has an after_load method, this method will be called
        E.g.:
        class Base(object):
            def after_load(self):
                self.app = self.pdriver.app

        Returns: bool True if a state was found; False if no state was found;
        """
        self.info_log("Loading state...")

        # Set the pdriver for an object that inherited from Stateful
        def set_pdriver(value):
            # TODO support class that inherited Stateful
            if Stateful in value.__class__.__bases__:
                value.pdriver = self.pdriver
                if hasattr(value, 'after_load'):
                    value.after_load()
            elif type(value) is list:
                for v in value:
                    set_pdriver(v)
            elif type(value) is dict:
                for k, v in iter(value.items()):
                    set_pdriver(v)

        # Load the state pickle
        state_pickle_path = self.get_state_pickle_path()
        if state_pickle_path:
            if os.path.isfile(state_pickle_path):
                with open(state_pickle_path, 'rb') as fd:
                    state = pickle.load(fd)
            else:
                self.info_log("No state found.")

                return False
        else:
            self.info_log("No state found.")
            return False

        # Set the pdriver
        for key, value in iter(state.items()):
            set_pdriver(value)

        # Update the instance variable dictionary
        self.__dict__.update(state)

        self.info_log("State loaded.")

        return True

    def configure_logger(self):
        """Configure the logger for the test

        Configure the logger from the logger_test config:
            logger_test:level
            logger_test:streamlogger
            logger_test:filelogger
            logger_test:format

        See documentation: http://brome.readthedocs.org/en/release/configuration.html#logger-test  # noqa

        If the config project:test_batch_result_path is set to False then the
            filelogger will not be configured.

        Returns: None
        """

        # Logger name
        logger_name = self._name

        # Log directory
        if self._runner_dir:
            self._test_log_dir = os.path.join(
                self._runner_dir,
                "logs"
            )
            self._relative_test_log_dir = os.path.join(
                self._runner.relative_runner_dir,
                "logs"
            )
        else:
            self._test_log_dir = ''
            self._relative_test_log_dir = ''

        # Logger
        self._logger = logging.getLogger(logger_name)

        # Create the log directory
        if self._runner_dir:
            create_dir_if_doesnt_exist(self._test_log_dir)

        # Format
        format_ = BROME_CONFIG['logger_test']['format']

        # Stream logger
        if BROME_CONFIG['logger_test']['streamlogger']:
            sh = logging.StreamHandler()
            stream_formatter = logging.Formatter(format_)
            sh.setFormatter(stream_formatter)
            self._logger.addHandler(sh)

        # File logger
        if self._runner_dir:
            if BROME_CONFIG['logger_test']['filelogger']:
                test_name = string_to_filename(self._name)
                self._log_file_path = os.path.join(
                    self._test_log_dir,
                    "%s_%s.log" %
                    (test_name, self._browser_config.get('browserName'))
                )
                self._relative_log_file_path = os.path.join(
                    self._relative_test_log_dir,
                    "%s_%s.log" %
                    (test_name, self._browser_config.get('browserName'))
                )
                fh = logging.FileHandler(
                    self._log_file_path
                )
                file_formatter = logging.Formatter(format_)
                fh.setFormatter(file_formatter)
                self._logger.addHandler(fh)

        # Set level
        self._logger.setLevel(getattr(
            logging, BROME_CONFIG['logger_test']['level'])
        )

    def get_logger_dict(self):
        return {
            'batchid': self._test_batch_friendly_name,
            'testname': u"%s" % self._name
        }

    def debug_log(self, msg):
        self._logger.debug(u"[debug]%s" % msg, extra=self.get_logger_dict())

    def info_log(self, msg):
        self._logger.info(u"%s" % msg, extra=self.get_logger_dict())

    def warning_log(self, msg):
        self._logger.warning(
            u"[warning]%s" % msg,
            extra=self.get_logger_dict()
        )

    def error_log(self, msg):
        self._logger.error(u"[error]%s" % msg, extra=self.get_logger_dict())

    def critical_log(self, msg):
        self._logger.critical(
            u"[critical]%s" % msg,
            extra=self.get_logger_dict()
        )

    def execute(self):
        try:
            self.before_run()

            if self._test_config.get("delete_state"):
                state_pickle_path = self.get_state_pickle_path()
                if state_pickle_path:
                    self.delete_state(state_pickle_path)
                else:
                    self.info_log('No state to delete!')

            if not self.load_state():
                if hasattr(self, 'create_state'):
                    ret = self.create_state()
                    if ret:
                        self.save_state()

            self.run(**self._test_config)

            self.after_run()

        except Exception as e:
            self.error_log('Test failed')

            tb = traceback.format_exc()

            self.error_log('Crash: %s' % tb)

            self.fail(tb)
        finally:
            self.end()

    def end(self):
        self.info_log("Test ended")

        # STOP PROXY
        if self._browser_config.location == 'ec2':
            instance = self._runner.resolve_instance_by_ip(self._private_ip)
            if instance.browser_config.get('enable_proxy'):
                instance.stop_proxy()

        self.stop_video_recording()
        ending_timestamp = utcnow()

        self.quit_driver()

        if BROME_CONFIG['runner']['play_sound_on_test_finished']:
            say(BROME_CONFIG['runner']['sound_on_test_finished'])

        ending_timestamp = ending_timestamp
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_instance = session.query(Testinstance)\
                .filter(Testinstance.mongo_id == self._test_instance_id).one()
            test_instance.ending_timestamp = ending_timestamp
            test_instance.terminated = True
            session.save(test_instance, safe=True)

        self.info_log("Ending timestamp: %s" % ending_timestamp)

    def quit_driver(self):
        self.info_log("Quitting the browser...")
        try:
            self.pdriver.quit()
        except Exception as e:
            self.error_log('Exception driver.quit(): %s' % str(e))

        """
        if hasattr(self, 'browsermobserver'):
            with open("temp.har", "w") as fd:
                fd.write(json.dumps(self.proxy.har))

            self.browsermobserver.stop()
        """

    def before_run(self):
        pass

    def after_run(self):
        pass

    def fail(self, tb):
        if BROME_CONFIG['runner']['play_sound_on_test_crash']:
            say(BROME_CONFIG['runner']['sound_on_test_crash'])

        if BROME_CONFIG['runner']['embed_on_test_crash']:
            self.pdriver.embed()

        if self.pdriver.bot_diary:
            if not self.pdriver.bot_diary.is_empty():
                traceback = tb
                tb = '\nBot Diary:\n'
                tb += self.pdriver.bot_diary.get_section(-1)[-1] + '\n'
                tb += 'And it failed\n'
                tb += str(traceback)

        self._crash_error = '[!]%s %s crashed: %s' % (self._name, self._browser_config.get_id(), str(tb))  # noqa

        self.create_crash_report(tb)

    def create_crash_report(self, tb):
        self.info_log('Creating a crash report')

        crash_name = "%s - %s" % (self.pdriver.get_id(join_char=' ', browser_version_join_char='.'), self._name)  # noqa

        extra_data = {}

        # JAVASCRIPT ERROR
        extra_data['javascript_error'] = self.pdriver.get_javascript_error()

        crash_screenshot_relative_path = ''
        if self._runner_dir:
            crash_screenshot_path = os.path.join(
                self._crash_report_dir,
                string_to_filename('%s.png' % crash_name)
            )

            crash_screenshot_relative_path = os.path.join(
                self._crash_report_relative_dir,
                string_to_filename('%s.png' % crash_name)
            )

            # CRASH LOG
            with open(os.path.join(self._crash_report_dir, string_to_filename('%s.log' % crash_name)), 'w') as f:  # noqa
                f.write(str(tb))

            # CRASH SCREENSHOT
            self.pdriver.take_screenshot(screenshot_path=crash_screenshot_path)

        # CRASH OBJECT
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            capabilities = {
                'browserName': self.pdriver.capabilities['browserName'],
                'platform': self.pdriver.capabilities['platform'],
                'version': self.pdriver.capabilities['version']
            }
            test_crash = Testcrash()
            test_crash.title = self._name
            test_crash.browser_capabilities = capabilities
            test_crash.browser_id = self.pdriver.get_id()
            test_crash.timestamp = utcnow()
            test_crash.trace = str(tb)
            if self._runner.root_test_result_dir:
                test_crash.root_path = self._runner.root_test_result_dir
                test_crash.screenshot_path = crash_screenshot_relative_path
                test_crash.video_capture_path = self._video_capture_file_relative_path  # noqa
            else:
                test_crash.root_path = ''
                test_crash.screenshot_path = ''
                test_crash.video_capture_path = ''
            test_crash.extra_data = extra_data
            test_crash.test_instance_id = self._test_instance_id
            test_crash.test_batch_id = self._test_batch_id

            session.save(test_crash, safe=True)

    def configure_test_result_dir(self):

        if not self._runner_dir:
            return

        # CRASH DIRECTORY
        self._crash_report_dir = os.path.join(
            self._runner_dir,
            'crashes'
        )
        create_dir_if_doesnt_exist(self._crash_report_dir)

        self._crash_report_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'crashes'
        )

        # ASSERTION SCREENSHOT DIRECTORY
        self._assertion_screenshot_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'assertion_screenshots',
            self.pdriver.get_id(join_char='_')
        )

        self._assertion_screenshot_dir = os.path.join(
            self._runner_dir,
            'assertion_screenshots',
            self.pdriver.get_id(join_char='_')
        )
        create_dir_if_doesnt_exist(self._assertion_screenshot_dir)

        # SCREENSHOT DIRECTORY
        self._screenshot_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'screenshots',
            self.pdriver.get_id(join_char='_')
        )

        self._screenshot_dir = os.path.join(
            self._runner_dir,
            'screenshots',
            self.pdriver.get_id(join_char='_')
        )
        create_dir_if_doesnt_exist(self._screenshot_dir)

        # QUALITY SCREENSHOT DIRECTORY
        self._quality_screenshot_relative_dir = os.path.join(
            self._runner.relative_runner_dir,
            'quality_screenshots',
            self.pdriver.get_id(join_char='_')
        )

        self._quality_screenshot_dir = os.path.join(
            self._runner_dir,
            'quality_screenshots',
            self.pdriver.get_id(join_char='_')
        )
        create_dir_if_doesnt_exist(self._quality_screenshot_dir)

        # VIDEO RECORDING DIRECTORY
        if self._browser_config.get('record_session'):
            self._video_recording_dir = os.path.join(
                self._runner_dir,
                'video_recording',
                self.pdriver.get_id(join_char='_')
            )
            create_dir_if_doesnt_exist(self._video_recording_dir)

            self._video_recording_relative_dir = os.path.join(
                self._runner.relative_runner_dir,
                'video_recording',
                self.pdriver.get_id(join_char='_')
            )

        # NETWORK CAPTURE DIRECTORY
        if self._browser_config.get('enable_proxy'):
            self._network_capture_dir = os.path.join(
                self._runner_dir,
                'network_capture'
            )
            create_dir_if_doesnt_exist(self._network_capture_dir)

    def get_test_result_summary(self):
        results = []

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            base_query = session.query(Testresult).filter(Testresult.test_instance_id == self._test_instance_id).filter(Testresult.browser_id == self.pdriver.get_id())  # noqa
            total_test = base_query.count()
            total_test_successful = base_query.filter(Testresult.result == True).count()  # noqa
            base_query = session.query(Testresult).filter(Testresult.test_instance_id == self._test_instance_id).filter(Testresult.browser_id == self.pdriver.get_id())  # noqa
            total_test_failed = base_query.filter(Testresult.result == False).count()  # noqa
            base_query = session.query(Testresult)\
                .filter(Testresult.test_instance_id == self._test_instance_id)\
                .filter(Testresult.browser_id == self.pdriver.get_id())
            failed_tests = base_query.filter(Testresult.result == False).all()  # noqa

            results.append('Total_test: %s; Total_test_successful: %s; Total_test_failed: %s' % (total_test, total_test_successful, total_test_failed))  # noqa

            failed_tests_title = []
            for failed_test in failed_tests:
                if failed_test.title not in failed_tests_title:
                    failed_tests_title.append(failed_test.title)
                    query = session.query(Test)\
                        .filter(Test.mongo_id == failed_test.test_id)
                    if query.count():
                        test = query.one()
                        test_id = test.test_id
                    else:
                        test_id = 'n/a'

                    results.append('[%s] %s' % (test_id, failed_test.title))

        return results



from time import sleep
from subprocess import Popen
from subprocess import call

import paramiko
import netifaces as ni

from brome.runner.base_instance import BaseInstance


class VirtualboxInstance(BaseInstance):
    """Virtual box instance

    Attributes:
        runner (object)
        browser_config (object)
        index (int)

    Kwargs:
        vbox (object)
    """

    def __init__(self, runner, browser_config, index, **kwargs):
        self.runner = runner
        self.browser_config = browser_config
        self.index = index

        self.vbox = kwargs.get('vbox')

    def get_ip(self):
        """Return the ip address of the node
        """

        return self.browser_config.get('ip')

    def execute_command(self, command, **kwargs):
        """Execute a command on the node

        Args:
            command (str)

        Kwargs:
            username (str)
        """

        self.info_log("executing command: %s" % command)

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            username = kwargs.get(
                'username',
                self.browser_config.get('username')
            )
            password = self.browser_config.get('password')

            ssh.connect(self.get_ip(), username=username, password=password)

            stdin, stdout, stderr = ssh.exec_command(command)

            ssh.close()

            return (stdout, stderr)

        except Exception as e:
            msg = "Execute_command exception: %s" % str(e)
            self.error_log(msg)
            raise Exception(msg)

    def scp_file_remote_to_local(self, remote_path, local_path):
        """Scp a remote file to local

        Args:
            remote_path (str)
            local_path (str)
        """

        sshadd_command = [
            'ssh-add',
            '/Users/pyrat/.ssh/ubuntuNode'
        ]
        self.info_log(
            "executing command: %s" %
            ' '.join(sshadd_command)
        )
        p = Popen(sshadd_command)
        p.wait()

        scp_command = [
            'scp',
            '-o',
            'StrictHostKeyChecking=no',
            '%s@%s:"%s"' %
            (
                self.browser_config.get('username'),
                self.get_ip(),
                remote_path
            ),
            local_path
        ]
        self.info_log(
            "executing command: %s" %
            ' '.join(scp_command)
        )
        p = Popen(scp_command)
        p.wait()

    def startup(self):
        """This will launch and configure the virtual box machine
        """

        import virtualbox

        # Do not launch the virtual machine
        if not self.browser_config.get('launch', False):
            return True

        self.info_log("Starting up")

        try:
            # Cleanup the vbox guestproperty variable
            call([
                'VBoxManage',
                'guestproperty',
                'delete',
                self.browser_config.get('vbname'),
                'wait_until_ready'
            ])
            call([
                'VBoxManage',
                'guestproperty',
                'delete',
                self.browser_config.get('vbname'),
                'hub_ip'
            ])

            # Find the machine
            vm = self.vbox.find_machine(self.browser_config.get('vbname'))

            # Session
            self.session = virtualbox.Session()

            # Determine if the vm is running or not
            # Launch it is necessary
            vm_already_running = False
            if str(vm.session_state) != 'Locked':
                vm.launch_vm_process(
                    self.session, self.browser_config.get(
                        'vbox_type', 'gui'
                    ), ''
                )
            else:
                vm_already_running = True

            user_session = vm.create_session()
            # NOTE this is another way to record the session
            # user_session.machine.video_capture_enabled = False

            # User session
            self.vbox_session = user_session

            # Wait until the vm is ready
            if not vm_already_running:

                instance_ready = False
                # TODO should be configurable
                timeout = 60

                self.info_log('Waiting for instance to start...')

                for i in range(timeout):
                    try:
                        value, timestamp, flags = user_session.machine.get_guest_property('wait_until_ready')  # noqa

                        if value == u'ready':
                            instance_ready = True
                            break

                    except virtualbox.library.VBoxErrorInvalidVmState:
                        pass

                    sleep(1)

                sleep(3)
                if instance_ready:
                    self.info_log('[Done] Instance ready...')
                else:
                    raise Exception("Timeout error: the virtualbox machine is still not ready.")  # noqa

            # HUB IP
            hub_ip = ni.ifaddresses('en0')[2][0]['addr']

            self.info_log("Hub ip: %s" % hub_ip)

            # VM IP
            vm_ip = vm.get_guest_property(
                "/VirtualBox/GuestInfo/Net/0/V4/IP"
            )[0]
            self.ip = vm_ip

            self.info_log("vm ip: %s" % vm_ip)

            # Start selenium on the node
            if not vm_already_running:

                # LINUX
                if self.browser_config.get('platform').lower() == "linux":

                    self.info_log('Starting the selenium node server')

                    # Update the hub_ip browser config
                    self.browser_config.config['hub_ip'] = hub_ip

                    command = self.browser_config.get(
                        "selenium_command"
                    ).format(**self.browser_config.config)
                    self.execute_command(command)

                # WINDOWS
                elif self.browser_config.get('platform').lower() == "windows":

                    self.info_log("Setting the guest property in Windows")

                    user_session.machine.set_guest_property(
                        "hub_ip", "%s:%s" % (hub_ip, '4444'), ''
                    )

            return True

        except Exception as e:
            self.error_log('Exception: %s' % e)
            raise

    def tear_down(self):
        """Tear down the virtual box machine
        """

        if not self.browser_config.get('terminate'):
            self.warning_log("Skipping terminate")
            return

        self.info_log("Tearing down")

        if self.browser_config.get('platform').lower() == 'linux':
            self.execute_command("shutdown -h now", username='root')

        elif self.browser_config.get('platform').lower() == 'windows':
            self.session.console.power_down()

    def start_video_recording(self, local_video_file_path, video_filename):
        """Start the video recording
        """

        self.runner.info_log("Starting video recording...")

        self.local_video_recording_file_path = local_video_file_path
        self.remote_video_recording_file_path = video_filename

        self.execute_command(
            "./start_recording.sh '%s'" % self.remote_video_recording_file_path
        )

    def stop_video_recording(self):
        """Stop the video recording
        """

        self.runner.info_log("Stopping video recording...")

        self.execute_command("./stop_recording.sh")
        # self.runner.info_log("output: %s"%output)

        sleep(5)

        self.scp_file_remote_to_local(
            self.remote_video_recording_file_path,
            self.local_video_recording_file_path
        )

    def get_id(self):
        return '%s - %s' % (self.browser_config.browser_id, self.index)

    def debug_log(self, msg):
        self.runner.debug_log("[%s]%s" % (self.get_id(), msg))

    def info_log(self, msg):
        self.runner.info_log("[%s]%s" % (self.get_id(), msg))

    def warning_log(self, msg):
        self.runner.warning_log("[%s]%s" % (self.get_id(), msg))

    def error_log(self, msg):
        self.runner.error_log("[%s]%s" % (self.get_id(), msg))

    def critial_log(self, msg):
        self.runner.critial_log("[%s]%s" % (self.get_id(), msg))



from brome.webserver.server.public.views import (
    index,
    api_get_session
)
from brome.webserver.server.auth.views import (
    Logout,
    Login,
    Register

)
from brome.webserver.server.crud.views import CRUD
from brome.webserver.server.brome_api.views import (
    LogStreamOut,
    GetBromeConfig,
    StartTestBatch
)

routes = [
    # CLIENT ROUTE => not /api/* and not /static/*
    ('GET', r'/{to:(?!api)(?!static)(?!tb_results).*}', index, 'index'),

    # API ROUTES
    ('POST', '/api/get_session', api_get_session, 'get_session'),
    ('*', '/api/crud', CRUD, 'api_crud'),
    ('*', '/api/getbromeconfig', GetBromeConfig, 'api_get_brome_config'),
    ('*', '/api/starttestbatch', StartTestBatch, 'api_start_test_batch'),
    ('*', '/api/logstreamout', LogStreamOut, 'api_log_stream_out'),
    ('*', '/api/login', Login, 'api_login'),
    ('*', '/api/register', Register, 'api_register'),
    ('*', '/api/logout', Logout, 'api_logout'),
]



import string
import os
import pytz
from datetime import datetime
import sys
from subprocess import call
from dateutil import tz

from redis import StrictRedis
from contextlib import ContextDecorator
from mongoalchemy.session import Session
from pymongo import MongoClient

import psutil


class DbSessionContext(ContextDecorator):
    def __init__(self, mongo_database_name):
        self.session = Session.connect(
            mongo_database_name,
            timezone=tz.gettz('UTC')
        )

    def __enter__(self):
        return self.session

    def __exit__(self, *exc):
        self.session.end()
        self.session.db.client.close()
        return False


def convert_tz_datetime(datetime_input, to_tz):
    return datetime_input.astimezone(tz.gettz(to_tz))


def utcnow():
    return datetime.now(tz=pytz.utc)


def get_timestamp():
    return datetime.now().strftime('%Y-%m-%d-%H-%M-%S')


def string_to_filename(s):
    valid_chars = "-_. %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in s if c in valid_chars)
    filename = filename.replace(' ', '_')
    return filename


def say(msg):
    if sys.platform in ['win32', 'linux2']:
        call(["espeak", msg])
    else:
        call(["say", msg])


def create_dir_if_doesnt_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)

if sys.platform == 'win32':
    devnull = open('log-null', 'w')
else:
    devnull = open('/dev/null', 'w')


def kill_by_pid(pid):
    p = psutil.Process(pid)
    p.terminate()
    print('[pid:%s]killed' % pid)


def kill_by_name(procname):
    for proc in psutil.process_iter():
        if proc.name() == procname:
            print('[pid:%s][name:%s] killed' % (proc.pid, proc.name()))
            proc.kill()


def kill_by_found_string_in_cmdline(procname, string):
    for proc in psutil.process_iter():
        if proc.name() == procname:
            for cmd in proc.cmdline():
                if cmd.find(string) != -1:
                    print('[pid:%s][name:%s] killed' % (proc.pid, proc.name()))
                    proc.kill()


def delete_database(db_name, flush_redis=True):
    mongo_client = MongoClient()
    mongo_client.drop_database(db_name)
    print('Database (%s) deleted!' % db_name)

    if flush_redis:
        redis_client = StrictRedis()
        redis_client.flushall()
        print('Redis flushed all')


def update_test(session, test_dict):
    from brome.model.test import Test
    print('Updating the test')
    for test_id, test_config in iter(test_dict.items()):
        if type(test_config) == dict:
            name = test_config['name']
        else:
            name = test_config

        # NEW TEST
        if not session.query(Test).filter(Test.test_id == test_id).count():
            test = Test(test_id=test_id, name=name)
            print('Added test id', test_id)
        # UPDATE TEST
        else:
            test = session.query(Test).filter(Test.test_id == test_id).one()
            if test.name != name:
                test.name = name
                print('Updated test id', test_id)

        session.save(test, safe=True)

    print('Done!')



import logging

from brome.core.utils import *
from brome.core.selector import Selector

class BotDiary(object):

    def __init__(self, pdriver):
        self.pdriver = pdriver
        self.sections = []
        self.section_index = -1
        self.entry_id = 0

        self.auto_diary = self.pdriver.get_config_value("bot_diary:enable_auto_bot_diary")

        self.configure_logger()

        self.add_section()

    def configure_logger(self):
        #Logger name
        logger_name = 'Bot Diary %s'%self.pdriver.test_instance._name

        #Log directory
        if self.pdriver.test_instance._runner_dir:
            self.bot_diary_dir = os.path.join(
                self.pdriver.test_instance._runner_dir,
                "bot_diary"
            )

            self.nodes_dir = os.path.join(
                self.pdriver.test_instance._runner_dir,
                "bot_diary",
                string_to_filename(self.pdriver.test_instance._name)
            )

            #Create the bot diary directory
            create_dir_if_doesnt_exist(self.bot_diary_dir)
            #Create the nodes screenshot directory
            create_dir_if_doesnt_exist(self.nodes_dir)

        #Logger
        self._logger = logging.getLogger(logger_name)

        #Stream logger 
        if self.pdriver.get_config_value('bot_diary:streamlogger'):
            sh = logging.StreamHandler()
            self._logger.addHandler(sh)

        #File logger
        if self.pdriver.test_instance._runner_dir:
            if self.pdriver.get_config_value('bot_diary:filelogger'):
                test_name = string_to_filename(self.pdriver.test_instance._name)
                fh = logging.FileHandler(os.path.join(
                    self.bot_diary_dir,
                    "%s.log"%test_name
                ))
                self._logger.addHandler(fh)

        self._logger.setLevel("INFO")

    def gen_next_entry_id(self):
        self.entry_id += 1

    def add_section(self):
        self.sections.append([])
        self.section_index += 1
        self._logger.info("[%d] ===%d==="%(self.get_next_entry_id(), self.section_index))
        self.gen_next_entry_id()

    def add_entry(self, entry):
        formatted_entry = "[%d] %s"%(self.get_next_entry_id(), entry)
        self._logger.info("%s"%formatted_entry)
        self.sections[self.section_index].append(formatted_entry)
        self.gen_next_entry_id()

    def add_auto_entry(self, action, **kwargs):
        """

        Keyword arguments:
            selector (str): default: false
            target (str): default: false
            take_screenshot (bool): default: false
        """

        selector = kwargs.get('selector', False)
        target = kwargs.get('target', False)
        take_screenshot = kwargs.get('take_screenshot', False)

        if target:
            effective_target = target
        elif selector:
            _selector = Selector(self.pdriver, selector)
            effective_target =  _selector.get_human_readable()
        else:
            effective_target = ""

        text_dict = {
            'action': action,
            'target': effective_target
        }

        if self.pdriver.get_config_value("bot_diary:enable_auto_bot_diary_component_screenshot") \
            and selector:
            self.pdriver.take_node_screenshot(
                self.pdriver.find(selector),
                os.path.join(self.nodes_dir, "%s.png"%self.get_next_entry_id())
            )
        elif take_screenshot:
            self.pdriver.take_screenshot(screenshot_path = os.path.join(self.nodes_dir, "%s.png"%self.get_next_entry_id()))

        if self.is_auto_diary_active():
            self.add_entry("{action} {target}.".format(**text_dict))
            return True
        else:
            return False

    def print_diary(self):
        for i, section in enumerate(self.sections):
            print('Section %d'%i)
            for entry in section:
                print(entry)

################################################################################ 
# IS
################################################################################ 

    def is_auto_diary_active(self):
        return self.auto_diary

    def is_empty(self):
        return self.sections[0] == []

################################################################################ 
# SET
################################################################################ 

    def set_enable_auto_diary(self):
        self.auto_diary = True

    def set_disable_auto_diary(self):
        self.auto_diary = False

    def set_section_index(self, index):
        self.section_index = index

################################################################################ 
# GET
################################################################################ 

    def get_next_entry_id(self):
        return self.entry_id + 1

    def get_section(self, index, join = False):
        if join:
            return '\n'.join(self.sections[index])
        else:
            return self.sections[index]

    def get_sections(self):
        return self.sections



import re
from io import StringIO

from lxml import etree
from lxml.cssselect import CSSSelector
from cssselect.parser import SelectorSyntaxError

from brome.core.settings import BROME_CONFIG


class Selector(object):
    """Selector class to manage the selector brome system

    Args:
        pdriver (object)
        selector (str)
    """
    def __init__(self, pdriver, selector):
        self._pdriver = pdriver
        self._selector = selector

        # List of selector support
        if not type(selector) is list:
            self._selector_list = [self._selector]
        else:
            self._selector_list = self._selector

        self._effective_selector_list = self.resolve_selector()
        self._effective_selector_type = self.resolve_selector_type()

        self.find_function, self.find_by = self.resolve_function()

        self._effective_selector = ''.join(
            [self._get_selector(sel) for sel in self._effective_selector_list]
        )

        # XPATH VALIDATION
        if self._effective_selector_type == 'xpath':
            if BROME_CONFIG['proxy_driver']['validate_xpath_selector']:
                xpath_test = etree.parse(StringIO('<foo><bar></bar></foo>'))
                try:
                    xpath_test.xpath(self.get_selector())
                except etree.XPathEvalError as e:
                    raise Exception("Invalid xpath: %s" % self.get_selector())

        # CSS VALIDATION
        elif self._effective_selector_type == 'css':
            if BROME_CONFIG['proxy_driver']['validate_css_selector']:
                try:
                    CSSSelector(self.get_selector())
                except SelectorSyntaxError:
                    raise Exception(
                        "Invalid css selector: %s" %
                        self.get_selector()
                    )

    def __repr__(self):
        return "Selector: effective selector (%s), raw selector (%s)" %\
            (self.get_selector(), self._selector)

    # GET
    def get_selector(self):
        """Get the final selector

        This selector can be feed to selenium
        """
        return self._effective_selector

    def get_human_readable(self):
        if len(self._selector_list) == 1:
            if self.get_type(self._selector_list[0]) == 'selector_variable':
                selector_config = BROME_CONFIG['selector_dict'][self._selector[3:]]  # noqa
                if type(selector_config) == dict:
                    if 'hr' in selector_config:
                        return selector_config['hr']
                    else:
                        return selector_config['default']
                else:
                    return selector_config

        return self.get_selector()

    def _get_selector(self, selector):
        """Get the selector without his brome specific prefix
        """
        return selector[3:]

    def get_type(self, selector):
        """Get the type of the selector

        see SELECTOR_DICT for supported type
        """
        try:
            return SELECTOR_DICT[selector[:3]]
        except KeyError:
            raise Exception("""
                All selector need to start with either:
                    'nm:' (name), 'xp:' (xpath), 'cn:' (classname), 'id:' (id),
                    'cs:' (css), 'tn:' (tag name), 'lt:' (link text),
                    'pl:' (partial link text)
                selector = %s
            """ % selector)

    # RESOLVE
    def resolve_selector(self):
        """Resolve the selector variable in place
        """
        effective_selector_list = []

        for current_selector in self._selector_list:

            # INLINE SELECTOR
            if self.get_type(current_selector) != 'selector_variable':
                effective_selector_list.append(current_selector)

            # SELECTOR VARIABLE
            else:
                # Make sure the proxy driver have a selector dictionary
                if self.get_type(current_selector) == 'selector_variable':
                    if not BROME_CONFIG['selector_dict']:
                        raise Exception("""
                            You must provide a selector dictionary if you want
                            to use the selector variable type
                        """)

                # Make sure that the selector dictionary
                # contains the selector variable
                if self._get_selector(current_selector) \
                        not in BROME_CONFIG['selector_dict']:
                    raise Exception("""
                        Cannot find the selector variable (%s)
                        in the selector dictionary
                    """ % self._get_selector(current_selector))

                effective_selector = BROME_CONFIG['selector_dict'][self._get_selector(current_selector)]  # noqa
                if type(effective_selector) is dict:
                    current_browser_id = False

                    keys = [key for key in effective_selector.keys()
                            if key not in ['default', 'hr']]

                    for key in keys:
                        for target in key.split('|'):
                            try:
                                re.search(
                                    target.lower(), self._pdriver.get_id().lower()
                                ).group(0)
                                current_browser_id = key
                            except AttributeError:
                                pass

                    if current_browser_id:
                        effective_selector_list.append(
                            effective_selector.get(current_browser_id)
                        )
                    else:
                        effective_selector_list.append(
                            effective_selector.get('default')
                        )

                else:
                    if self.get_type(effective_selector) in \
                            [value for key, value in SELECTOR_DICT.items()
                                if key != 'selector_variable']:
                        effective_selector_list.append(effective_selector)
                    else:
                        raise Exception("""
                            All selector need to start with either:
                                'nm:' (name), 'xp:' (xpath), 'cn:' (classname),
                                'id:' (id), 'cs:' (css), 'tn:' (tag name),
                                'lt:' (link text), 'pl:' (partial link text)
                        """)

        return effective_selector_list

    def resolve_selector_type(self):
        """Resolve the selector type

        This make sure that all the selectors
        provided are of the same type (in case of a list of selector)
        """
        resolved_selector_type_list = []
        for current_selector in self._effective_selector_list:
            resolved_selector_type_list.append(self.get_type(current_selector))

        set_ = set(resolved_selector_type_list)

        # VALIDATE THAT ALL SELECTOR ARE OF THE SAME TYPE
        if len(set_) != 1:
            raise Exception("""
                If you provide a list of selector then
                all selectors must be of the same type
            """)
        else:
            return set_.pop()

    def resolve_function(self):
        """Resolve the selenium function that will be use to find the element
        """
        selector_type = self._effective_selector_type

        # NAME
        if selector_type == 'name':
            return ('find_elements_by_name', 'NAME')

        # XPATH
        elif selector_type == 'xpath':
            return ('find_elements_by_xpath', 'XPATH')

        # CLASSNAME
        elif selector_type == 'class_name':
            return ('find_elements_by_class_name', 'CLASS_NAME')

        # ID
        elif selector_type == 'id':
            return ('find_element_by_id', 'ID')

        # CSS
        elif selector_type == 'css':
            return ('find_elements_by_css_selector', 'CSS_SELECTOR')

        # TAGNAME
        elif selector_type == 'tag_name':
            return ('find_elements_by_tag_name', 'TAG_NAME')

        # LINK TEXT
        elif selector_type == 'link_text':
            return ('find_elements_by_link_text', 'LINK_TEXT')

        # PARTIAL LINK TEXT
        elif selector_type == 'partial_link_text':
            return ('find_elements_by_partial_link_text', 'PARTIAL_LINK_TEXT')

SELECTOR_DICT = {
    'nm:': 'name',
    'xp:': 'xpath',
    'lt:': 'link_text',
    'pl:': 'partial_link_text',
    'cn:': 'class_name',
    'id:': 'id',
    'cs:': 'css',
    'tn:': 'tag_name',
    'sv:': 'selector_variable'
}



import traceback
from time import sleep

from selenium.webdriver.common.touch_actions import TouchActions
from selenium.common import exceptions
from selenium.webdriver.common.action_chains import ActionChains

from brome.core.settings import BROME_CONFIG


class ProxyElement(object):

    def __init__(self, element, selector, pdriver):
        self._element = element
        self.selector = selector
        self.pdriver = pdriver

    def __getattr__(self, funcname):
        return getattr(self._element, funcname)

    def error_log(self, msg):
        self.pdriver.error_log("[%s] %s" % (self.__repr__(), msg))

    def debug_log(self, msg):
        self.pdriver.debug_log("[%s] %s" % (self.__repr__(), msg))

    def __repr__(self):
        msg = ["WebElement (selector: '%s')" % self.selector.get_selector()]

        try:
            if self._element.get_attribute('id'):
                msg.append("(id: '%s')" % self._element.get_attribute('id'))

            if self._element.get_attribute('name'):
                msg.append(
                    "(name: '%s')" % self._element.get_attribute('name')
                )

            if self._element.get_attribute('class'):
                msg.append(
                    "(class: '%s')" % self._element.get_attribute('class')
                )

        except Exception as e:
            self.pdriver.debug_log("exception in __repr__: %s" % e)

        return ' '.join(msg)

    def is_displayed(self, **kwargs):
        self.debug_log("Is displayed")

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        retry = kwargs.get('retry', True)

        try:
            is_displayed = self._element.is_displayed()
            self.debug_log("is displayed")
            return is_displayed
        except exceptions.StaleElementReferenceException:
            if retry:
                # NOTE this is an imperfect solution since we can
                # have found the element with find_last
                # TODO find a better way to handle this edge case
                self.debug_log("StaleElementReferenceException; retrying...")
                self._element = self.pdriver.find(
                    self.selector._selector,
                    raise_exception=False
                )
                if self._element:
                    return self.is_displayed(retry=False)
                else:
                    return False
            else:
                if raise_exception:
                    raise
                else:
                    return False

    def click(self, **kwargs):
        self.debug_log("Clicking on element")

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_when_element_is_clicked']
        )
        wait_until_clickable = kwargs.get(
            'wait_until_clickable',
            BROME_CONFIG['proxy_element']['wait_until_clickable']
        )

        if wait_until_clickable:
            # TODO manage the raise exception better
            self.pdriver.wait_until_clickable(
                self.selector._selector,
                raise_exception=True
            )

        if highlight:
            self.highlight(
                style=BROME_CONFIG['highlight']['style_when_element_is_clicked']  # noqa
            )

        def _click():
            if BROME_CONFIG['proxy_element']['use_touch_instead_of_click']:
                touch_action = TouchActions(self.pdriver._driver)
                touch_action.tap(self._element).perform()
            else:
                self._element.click()

        if self.pdriver.bot_diary:
            self.pdriver.bot_diary.add_auto_entry(
                "I clicked on",
                selector=self.selector._selector
            )

        try:
            _click()
        except (
                    exceptions.InvalidElementStateException,
                    exceptions.WebDriverException
                ) as e:
            self.debug_log("click exception: %s" % e)
            sleep(2)
            self.scroll_into_view()
            _click()
        except exceptions.StaleElementReferenceException as e:
            self.debug_log(
                "click exception StaleElementReferenceException: %s" % e
            )
            sleep(2)
            self._element = self.pdriver.find(self.selector._selector)
            _click()

        return True

    def double_click(self, **kwargs):
        self.debug_log("Double clicking on element")

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_when_element_is_clicked']
        )
        wait_until_clickable = kwargs.get(
            'wait_until_clickable',
            BROME_CONFIG['proxy_element']['wait_until_clickable']
        )

        if wait_until_clickable:
            # TODO manage the raise exception better
            self.pdriver.wait_until_clickable(
                self.selector._selector,
                raise_exception=True
            )

        if highlight:
            self.highlight(
                style=BROME_CONFIG['highlight']['style_when_element_is_clicked']  # noqa
            )

        def _double_click():
            if BROME_CONFIG['proxy_element']['use_touch_instead_of_click']:
                touch_action = TouchActions(self.pdriver._driver)
                touch_action.double_tap(self._element).perform()
            else:
                ActionChains(self.pdriver)\
                    .double_click(self._element).perform()

        if self.pdriver.bot_diary:
            self.pdriver.bot_diary.add_auto_entry(
                "I double clicked on",
                selector=self.selector._selector
            )

        try:
            _double_click()
        except (
                    exceptions.InvalidElementStateException,
                    exceptions.WebDriverException
                ) as e:
            self.debug_log("click exception: %s" % e)
            sleep(2)
            self.scroll_into_view()
            _double_click()
        except exceptions.StaleElementReferenceException as e:
            self.debug_log(
                "click exception StaleElementReferenceException: %s" % e
            )
            sleep(2)
            self._element = self.pdriver.find(self.selector._selector)
            _double_click()

        return True

    def send_keys(self, value, **kwargs):
        self.debug_log("Sending keys")

        highlight = kwargs.get(
            'highlight',
            BROME_CONFIG['highlight']['highlight_when_element_receive_keys']  # noqa
        )
        """
        wait_until_clickable = kwargs.get(
            'wait_until_clickable',
            BROME_CONFIG['proxy_element']['wait_until_clickable']
        )

        if wait_until_clickable:
            # TODO manage the raise exception better
            self.pdriver.wait_until_clickable(
                self.selector._selector,
                raise_exception=True
            )
        """

        if highlight:
            self.highlight(
                style=BROME_CONFIG['highlight']['style_when_element_receive_keys']  # noqa
            )

        clear = kwargs.get('clear', False)

        if clear:
            self.clear()

        if self.pdriver.bot_diary:
            self.pdriver.bot_diary.add_auto_entry(
                "I typed '%s' in" % value,
                selector=self.selector._selector
            )

        try:
            self._element.send_keys(value)
        except exceptions.StaleElementReferenceException as e:
            self.debug_log(
                "send_keys exception StaleElementReferenceException: %s" % e
            )
            sleep(2)
            self._element = self.pdriver.find(self.selector._selector)
            self._element.send_keys(value)
        except (
                exceptions.InvalidElementStateException,
                exceptions.WebDriverException
                ) as e:
            self.debug_log("send_keys exception: %s" % e)
            sleep(2)
            self._element.send_keys(value)

        return True

    def clear(self):
        self.debug_log("Clearing element")

        try:
            self._element.clear()
        except (
                exceptions.InvalidElementStateException,
                exceptions.WebDriverException
                ) as e:
            self.debug_log("send_keys exception: %s" % e)
            sleep(2)
            self._element.clear()
        except exceptions.StaleElementReferenceException as e:
            self.debug_log(
                "send_keys exception StaleElementReferenceException: %s" % e
            )
            sleep(2)
            self._element = self.pdriver.find(self.selector._selector)
            self._element.clear()

        return True

    def highlight(self, **kwargs):
        """
            kwargs:
                style: css
                highlight_time: int; default: .3
        """

        self.debug_log("Highlighting element")

        style = kwargs.get('style')
        highlight_time = kwargs.get('highlight_time', .3)

        driver = self._element._parent

        try:
            original_style = self._element.get_attribute('style')

            driver.execute_script(
                "arguments[0].setAttribute('style', arguments[1]);",
                self._element,
                style
            )
        except Exception as e:
            self.error_log("highlight exception: %s" % str(e))

        sleep(highlight_time)

        try:
            driver.execute_script(
                "arguments[0].setAttribute('style', arguments[1]);",
                self._element,
                original_style
            )
        except Exception as e:
            self.error_log("highlight exception: %s" % str(e))

        return True

    def scroll_into_view(self, **kwargs):
        self.debug_log("Scrolling into view element")

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )
        try:
            self.pdriver.execute_script(
                "arguments[0].scrollIntoView()",
                self._element
            )

        except (
                exceptions.WebDriverException,
                exceptions.StaleElementReferenceException
                ) as e:
            if raise_exception:
                raise
            else:
                tb = traceback.format_exc()
                self.error_log('scroll_into_view WebDriverException: %s' % tb)
                return False

        return True

    def select_all(self, **kwargs):
        self.debug_log("Selecting all in element")

        raise_exception = kwargs.get(
            'raise_exception',
            BROME_CONFIG['proxy_driver']['raise_exception']
        )

        # http://stackoverflow.com/questions/985272/selecting-text-in-an-element-akin-to-highlighting-with-your-mouse  # noqa
        try:
            self.pdriver.execute_script("""
                var element = arguments[0],
                    range, selection;

                if (document.body.createTextRange) {
                    range = document.body.createTextRange();
                    range.moveToElementText(element);
                    range.select();
                } else if (window.getSelection) {
                    selection = window.getSelection();
                    range = document.createRange();
                    range.selectNodeContents(element);
                    selection.removeAllRanges();
                    selection.addRange(range);
                }
            """, self._element)
        except exceptions.WebDriverException:
            if raise_exception:
                raise
            else:
                tb = traceback.format_exc()
                self.error_log('select_all WebDriverException: %s' % tb)
                return False

        return True

    def hover(self, **kwargs):
        self.debug_log("Hovering element")

        try:
            ActionChains(self.pdriver)\
                .move_to_element(self._element)\
                .perform()
        except exceptions.WebDriverException:
            tb = traceback.format_exc()
            self.error_log('hover WebDriverException: %s' % tb)
            return False

        return True



import pickle
import hashlib
import shutil
import os
import argparse
import re
from glob import glob


from brome.core.utils import (
    update_test,
    delete_database,
    DbSessionContext
)
from brome.webserver.server.app import run_app
from brome.core.grep import grep_files
from brome.runner.local_runner import LocalRunner
from brome.runner.grid_runner import GridRunner
from brome.core.settings import BROME_CONFIG
from brome.core.configurator import (
    generate_brome_config
)

__version__ = "1.1.1"


class Brome(object):
    def __init__(self, **kwargs):
        self.configure(**kwargs)

    def configure(self, **kwargs):
        brome_config = kwargs.get('config', dict())
        selector_dict = kwargs.get('selector_dict', dict())
        test_dict = kwargs.get('test_dict', dict())
        browsers_config = kwargs.get('browsers_config', dict())
        self.tests = kwargs.get('tests', list())

        if 'project' in brome_config:
            absolute_path = kwargs.get('absolute_path')
            brome_config['project']['absolute_path'] = absolute_path
            if not absolute_path:
                print('[Warning] absolute_path not provided in the brome.configure()')  # noqa

        BROME_CONFIG.update(generate_brome_config())
        for key in iter(BROME_CONFIG):
            if key in brome_config:
                BROME_CONFIG[key].update(brome_config[key])

        for key in iter(brome_config):
            if key not in BROME_CONFIG:
                BROME_CONFIG[key] = brome_config[key]

        BROME_CONFIG['selector_dict'] = selector_dict
        BROME_CONFIG['browsers_config'] = browsers_config
        BROME_CONFIG['test_dict'] = test_dict

    def print_usage(self):
        print('$ ./bro admin | run | webserver | list | find')
        exit(1)

    def execute(self, args):
        if len(args) < 2:
            self.print_usage()

        if args[1] == 'run':
            self.run(args[2:])
        elif args[1] == 'admin':
            self.admin(args[2:])
        elif args[1] == 'find':
            self.find(args[2:])
        elif args[1] == 'list':
            self.list_(args[2:])
        elif args[1] == 'webserver':
            self.webserver(args[2:])
        else:
            self.print_usage()

    def run(self, args):
        parser = argparse.ArgumentParser(description='Brome runner')

        # TEST SEARCH QUERY
        parser.add_argument(
                            '--search',
                            '-s',
                            dest='test_search_query',
                            default='*',
                            help='The search query used to activate tests'
        )
        # TEST NAME
        parser.add_argument(
                            '--name',
                            '-n',
                            dest='test_name',
                            help='The test name to be executed'
        )

        # TEST FILE
        parser.add_argument(
            '--test-file',
            dest='test_file',
            default=False,
            help='The file containing the tests name'
        )

        # LOCALHOST RUNNER
        parser.add_argument(
                            '--localhost',
                            '-l',
                            dest='localhost_runner',
                            help='Launch a browser from localhost'
        )

        # REMOTE RUNNER
        parser.add_argument(
                            '-r',
                            '--remote',
                            dest='remote_runner',
                            default=False,
                            help='Launch a browser on a remote'
        )

        def test_config_string(value):
            try:
                re.match("([^=]+=[^,]+,?)+", value).group(0)
                return value
            except AttributeError:
                print("--test-config '%s' does not match required format (key=value,key1:value1)" % (value))  # noqa
                exit(1)

        # TEST CONFIG KWARGS
        parser.add_argument(
                            '--test-config',
                            type=test_config_string,
                            dest='test_config',
                            help='The config that will be pass to the test. ex: "key=value,key1=value1"'  # noqa
        )

        parsed_args = parser.parse_args(args)
        run_args = vars(parsed_args)
        BROME_CONFIG['runner_args'] = run_args

        if BROME_CONFIG['test_dict']:
            self.auto_update_test()

        if run_args['localhost_runner']:
            browsers_id = [run_args['localhost_runner']]
        elif run_args['remote_runner']:
            browsers_id = run_args['remote_runner'].split(',')
        else:
            raise Exception('Please select -r or -l')

        for browser_id in browsers_id:
            if browser_id not in BROME_CONFIG['browsers_config'].keys():
                raise Exception("""
                    This browser id is not available
                    in the provided browsers config\n
                    Supported browser(s) are: {browsers}
                """.format(
                    browsers=BROME_CONFIG['browsers_config'].keys()
                ))

        # LOCALHOST RUNNER
        if run_args['localhost_runner']:
            LocalRunner(self).execute()

        # REMOTE RUNNER
        elif run_args['remote_runner']:
            GridRunner(self).execute()

        # ERROR
        else:
            raise Exception(
                'Select either -l "{browser_id}" or -r "{browser_id}"'
            )

    def webserver(self, args):
        run_app()

    def list_(self, args):
        query = os.path.join(
                        BROME_CONFIG['project']['absolute_path'],
                        BROME_CONFIG['brome']['script_folder_name'],
                        "%s*.py" %
                        BROME_CONFIG['brome']['script_test_prefix']
                    )

        tests = glob(query)
        print("[index]\t|test name|")
        for index, test in enumerate(tests):
            test_name = test.split(os.sep)[-1][5:-3]
            print("[%s]\t%s" % (index, test_name))

    def admin(self, args):
        parser = argparse.ArgumentParser(description='Brome admin')

        parser.add_argument(
                            '--reset',
                            dest='reset',
                            action='store_true',
                            help='delete the database + delete the test batch results + update the test collection'  # noqa
        )

        parser.add_argument(
                            '--delete-test-states',
                            dest='delete_test_states',
                            action='store_true',
                            help='Delete all the test states'
        )

        parser.add_argument(
                            '--delete-test-results',
                            dest='delete_test_result',
                            action='store_true',
                            help='Delete all the test batch results'
        )

        parser.add_argument(
                            '--delete-database',
                            dest='delete_database',
                            action='store_true',
                            help='Delete the project database'
        )

        parser.add_argument(
                            '--update-test',
                            dest='update_test',
                            action='store_true',
                            help='Update the test in the database'
        )

        parsed_args = parser.parse_args(args)

        def delete_test_states():
            states_dir = os.path.join(
                BROME_CONFIG["project"]["absolute_path"],
                BROME_CONFIG["brome"]["script_folder_name"],
                "states"
            )
            try:
                shutil.rmtree(states_dir)
            except OSError:
                pass

            try:
                os.makedirs(states_dir)
            except OSError:
                pass

            print('States deleted')

        def delete_test_batch_result():
            tb_results_path = os.path.join(
                BROME_CONFIG['project']['test_batch_result_path'],
                'tb_results'
            )
            if os.path.exists(tb_results_path):
                shutil.rmtree(tb_results_path)
                print('Test batch result (%s) deleted!' % tb_results_path)
            else:
                print('Nothing to delete')

        if parsed_args.reset:
            delete_database(BROME_CONFIG['database']['mongo_database_name'])
            delete_test_batch_result()
            if BROME_CONFIG['test_dict']:
                self.update_test()
        elif parsed_args.delete_test_result:
            delete_test_batch_result()
        elif parsed_args.delete_test_states:
            delete_test_states()
        elif parsed_args.delete_database:
            delete_database(BROME_CONFIG['database']['mongo_database_name'])
        elif parsed_args.update_test:
            self.update_test()

    def find(self, args):
        parser = argparse.ArgumentParser(description='Brome find')

        parser.add_argument(
                            '--test-id',
                            dest='test_id',
                            help='Find a test_id in model and test directory'
        )

        parser.add_argument(
                            '--unused-test-id',
                            dest='unused_test_id',
                            action='store_true',
                            help='Find all unused test id in model and test directory'  # noqa
        )

        parser.add_argument(
                            '--selector',
                            dest='selector',
                            help='Find a selector in model and test directory'
        )

        parser.add_argument(
                            '--unused-selector',
                            dest='unused_selector',
                            action='store_true',
                            help='Find all unused selector variable in model and test directory'  # noqa
        )

        parsed_args = parser.parse_args(args)

        test_id_regex = "[\'\"]+(%s){1}[\'\"]+"
        selector_regex = "\..*\(+[\'\"]+.*(%s)+.*[\'\"]+"
        if parsed_args.test_id:
            pattern = test_id_regex % parsed_args.test_id
        elif parsed_args.selector:
            pattern = selector_regex % parsed_args.selector

        paths = [
            os.path.join(
                BROME_CONFIG['project']['absolute_path'],
                BROME_CONFIG['brome']['script_folder_name']
            ),
            os.path.join(
                BROME_CONFIG['project']['absolute_path'],
                "model"
            ),
        ]

        if parsed_args.test_id or parsed_args.selector:
            result = grep_files(paths, pattern, True)
        else:
            if parsed_args.unused_selector:
                print('Unused selector variable:')
                dict_ = BROME_CONFIG['selector_dict']
                regex_ = selector_regex

            elif parsed_args.unused_test_id:
                print('Unused test id:')
                dict_ = BROME_CONFIG['test_dict']
                regex_ = test_id_regex

            try:
                for key in dict_:
                    pattern = regex_ % key
                    result = grep_files(paths, pattern, True, return_list=True)
                    if not result:
                        print(key)

            except KeyboardInterrupt:
                pass

    def update_test(self):
            if BROME_CONFIG['test_dict']:
                with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                    update_test(session, BROME_CONFIG['test_dict'])
            else:
                raise Exception("No test dictionary provided")

    def auto_update_test(self):
        brome_memory_pickle_path = os.path.join(
            BROME_CONFIG['project']['absolute_path'],
            ".brome.pkl"
        )

        new_hash = hashlib.sha1(
            BROME_CONFIG['test_dict'].__repr__().encode('utf')
        ).hexdigest()

        if os.path.isfile(brome_memory_pickle_path):
            with open(brome_memory_pickle_path, "rb") as fd:
                data = pickle.load(fd)

            if 'test_dict_hash' in data:
                if data['test_dict_hash'] != new_hash:
                    self.update_test()

                    data['test_dict_hash'] = new_hash
                    with open(brome_memory_pickle_path, "wb") as fd:
                        pickle.dump(data, fd)
        else:
            data = {}
            data['test_dict_hash'] = new_hash

            with open(brome_memory_pickle_path, "wb") as fd:
                pickle.dump(data, fd)
            self.update_test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = []
with open('requirements.txt', 'r') as fd:
    for line in fd:
        requirements.append(line)

setup(
    name='brome',
    version='1.1.1',
    description="Selenium Framework",
    long_description=readme,
    author="Jean-Francois Parent",
    author_email='parent.j.f@gmail.com',
    url='https://github.com/jf-parent/brome',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords=['brome', 'selenium'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3.5",
    ]
)



import socket
import os
from subprocess import Popen

import psutil

from brome.core.settings import BROME_CONFIG
from brome.core.utils import (
    create_dir_if_doesnt_exist,
    string_to_filename,
    kill_by_pid
)
from brome.runner.base_instance import BaseInstance


class LocalhostInstance(BaseInstance):
    """Localhost instance

    Attributes:
        runner (object)
        browser_config (object)
    """

    def __init__(self, runner, browser_config, **kwargs):
        self.runner = runner
        self.browser_config = browser_config
        self.test_name = kwargs.get('test_name')

    def startup(self):
        """Start the instance

        This is mainly use to start the proxy
        """
        self.runner.info_log("Startup")

        if self.browser_config.config.get('enable_proxy'):
            self.start_proxy()

    def tear_down(self):
        """Tear down the instance

        This is mainly use to stop the proxy
        """
        self.runner.info_log("Tear down")

        if self.browser_config.config.get('enable_proxy'):
            self.stop_proxy()

    def execute_command(self, command):
        """Execute a command

        Args:
            command (str)

        Returns:
            process (object)
        """

        self.runner.info_log("Executing command: %s" % command)

        process = Popen(
                command,
                stdout=open(os.devnull, 'w'),
                stderr=open('runner.log', 'a'),
        )

        return process

    def start_proxy(self, port=None):
        """Start the mitmproxy
        """

        self.runner.info_log("Starting proxy...")

        # Get a random port that is available
        if not port:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(('0.0.0.0', 0))
            sock.listen(5)
            self.proxy_port = sock.getsockname()[1]
            sock.close()

        network_data_path = os.path.join(
            self.runner.runner_dir,
            'network_capture'
        )
        create_dir_if_doesnt_exist(network_data_path)

        self.proxy_output_path = os.path.join(
            network_data_path,
            string_to_filename('%s.data' % self.test_name)
        )

        path_to_mitmproxy = BROME_CONFIG['mitmproxy']['path']

        if not path_to_mitmproxy:
            raise Exception("""
                You need to set the mitmproxy:path config to be able
                to the use the proxy with this browser
            """)

        filter_ = BROME_CONFIG['mitmproxy']['filter']
        command = [
            path_to_mitmproxy,
            "-p",
            "%s" % self.proxy_port,
            "-w",
            self.proxy_output_path
        ]

        if filter_:
            command.append(filter_)

        process = self.execute_command(command)

        self.proxy_pid = process.pid

        self.runner.info_log("Proxy pid: %s" % self.proxy_pid)

    def stop_proxy(self):
        """Stop the mitmproxy
        """

        self.runner.info_log("Stopping proxy...")

        if hasattr(self, 'proxy_pid'):
            try:
                kill_by_pid(self.proxy_pid)
            except psutil.NoSuchProcess:
                pass




import os
import binascii

import jinja2
from mongoalchemy.fields import StringField


def generate_token(n):
    return binascii.hexlify(os.urandom(n)).decode('utf')


class SafeStringField(StringField):

    def __set__(self, instance, value):
        value = jinja2.utils.escape(value)
        self.set_value(instance, value)
